# -*- coding: utf-8 -*-
"""
qspec._simulate_cpp
===================

Classes and methods for the 'simulate' module using the Python/C++ interface.
"""

import numpy as np
from matplotlib import cm
import matplotlib.pyplot as plt

from qspec.qtypes import *
from qspec._cpp import *
from qspec import tools
from qspec import get_f, get_m
import qspec.algebra as al


__all__ = ['Polarization', 'Laser', 'Environment', 'construct_electronic_state', 'construct_hyperfine_state', 'State',
           'DecayMap', 'Atom', 'Interaction']


def sr_generate_y(denominator: np.ndarray, f_theta: np.ndarray, f_phi: np.ndarray,
                  counts: np.ndarray, shape: np.ndarray):
    """
    :param denominator: The denominator of the scattering rate.
    :param f_theta: The numerator with the 'theta-polarization'.
    :param f_phi: The numerator with the 'phi-polarization'.
    :param counts: The number of summands.
    :param shape: The shape of y.
    :returns: The scattering rate.
    """
    y = np.zeros((shape[0] * shape[1], ), dtype=float)  # Allocate memory.
    denominator_p = denominator.ctypes.data_as(c_complex_p)  # Get all pointers to the first elements of the arrays.
    f_theta_p = f_theta.ctypes.data_as(c_complex_p)
    f_phi_p = f_phi.ctypes.data_as(c_complex_p)
    counts_p = counts.ctypes.data_as(c_size_t_p)
    shape_p = shape.ctypes.data_as(c_size_t_p)
    y_p = y.ctypes.data_as(c_double_p)
    dll.sr_generate_y(denominator_p, f_theta_p, f_phi_p, counts_p, shape_p, y_p)  # Modify y "in-place" with C++.
    return y


def _process_q_axis(q_axis: array_like) -> ndarray:
    """
    Preprocess the quantization axis.

    :param q_axis: The quantization axis. Must be an integer in {0, 1, 2} or a 3d-vector.
    :returns: The quantization axis as a 3d-vector.
    """
    q_axis = np.asarray(q_axis, dtype=float)
    if not q_axis.shape:
        if q_axis % 1 != 0:
            raise ValueError('q_axis must be an integer or a 3d-vector.')
        elif int(q_axis) not in {0, 1, 2}:
            raise ValueError('q_axis must be in {0, 1, 2}.')
        else:
            q_axis = tools.unit_vector(int(q_axis), 3)
    elif q_axis.shape != (3,):
        raise ValueError('q_axis must be an integer or a 3d-vector.')
    return q_axis


class Polarization:
    def __init__(self, vec: array_iter = None, q_axis: array_like = 2, vec_as_q: bool = True, instance=None):
        """
        Class representing a polarization state of light. The property 'x' holds the polarization in cartesian coordinates.
        The property 'q' holds the polarization in spherical coordinates ( sigma-, pi, sigma+ )
        with respect to the chosen quantization axis.

        :param vec: The polarization vector. I.e. the amplitude of the electromagnetic wave. So to specify, e.g.,
         1/3 of pi and 2/3 sigma+ light for a given 'q_axis', vec must be ( 0, sqrt(1/3), sqrt(2/3) ).
         The default is linear polarization in z-direction, such that x = (0, 0, 1) and q = (0, 1, 0).
        :param q_axis: The quantization axis. Must be an integer in {0, 1, 2} or a 3d-vector. The default is 2 (z-axis).
        :param vec_as_q: Whether 'vec' is given as ( sigma-, pi, sigma+ ) (True) or in cartesian coordinates (False).
        :param instance: A pointer to an existing Polarization instance.
         If this is specified, the other parameters are omitted.
        """
        self.instance = instance
        if self.instance is None:
            self.instance = dll.polarization_construct()

            if vec is None:
                vec = tools.unit_vector(1, 3, dtype=complex)
            vec = np.asarray(vec, dtype=complex)
            if vec.shape != (3,):
                raise ValueError('vec must be a 3d-vector.')

            q_axis = _process_q_axis(q_axis)
            dll.polarization_init(self.instance, vec, q_axis, vec_as_q)

    def __del__(self):
        dll.polarization_destruct(self.instance)

    def def_q_axis(self, q_axis: array_like = 2, q_fixed: bool = False):
        """
        Defines the quantization axis. This changes either 'x' or 'q', depending on 'q_fixed'.

        :param q_axis: The quantization axis. Must be an integer in {0, 1, 2} or a 3d-vector. The default is 2 (z-axis).
        :param q_fixed: Whether 'q' should stay the same with the new quantization axis (True) or 'x' (False).
        """
        q_axis = _process_q_axis(q_axis)
        dll.polarization_def_q_axis(self.instance, q_axis, q_fixed)

    @property
    def q_axis(self) -> ndarray:
        """
        :returns: The complex polarization in spherical coordinates ( sigma-, pi, sigma+ ).
        """
        return dll.polarization_get_q_axis(self.instance)

    @property
    def x(self) -> ndarray:
        """
        :returns: The complex polarization in cartesian coordinates ( x, y, z ).
        """
        return dll.polarization_get_x(self.instance)

    @property
    def q(self) -> ndarray:
        """
        :returns: The complex polarization in spherical coordinates ( sigma-, pi, sigma+ ).
        """
        return dll.polarization_get_q(self.instance)


class Laser:
    def __init__(self, freq: scalar, intensity: scalar = 1., polarization: Polarization = None,
                 k: array_like = None, instance=None):
        """
        Class representing a laser.

        :param freq: The frequency of the laser (MHz).
        :param intensity: The intensity of the laser (uW / mm**2 = W / m**2).
        :param polarization: The polarization of the laser.
        :param k: The direction of the laser.
        :param instance: A pointer to an existing Laser instance.
         If this is specified, the other parameters are omitted.
        """
        self.instance = instance
        if self.instance is None:
            self.instance = dll.laser_construct()
            self._polarization = polarization
            if self.polarization is None:
                self._polarization = Polarization()
            if k is None:
                k = tools.unit_vector(0, 3)
            k = np.asarray(k, dtype=float)
            if k.shape != (3,):
                raise ValueError('Interaction.k must be a 3d-vector, but has shape {}.'.format(k.shape))
            dll.laser_init(self.instance, c_double(freq), c_double(intensity), polarization.instance,
                           k.ctypes.data_as(c_double_p))
        else:
            self._polarization = Polarization(instance=dll.laser_get_polarization(self.instance))

    def __del__(self):
        dll.laser_destruct(self.instance)

    @property
    def freq(self):
        """
        :returns: The frequency of the laser.
        """
        return dll.laser_get_freq(self.instance)

    @freq.setter
    def freq(self, value: scalar):
        """
        :param value: The new frequency of the laser.
        :returns: None.
        """
        dll.laser_set_freq(self.instance, c_double(value))

    @property
    def intensity(self):
        """
        :returns: The intensity of the laser.
        """
        return dll.laser_get_intensity(self.instance)

    @intensity.setter
    def intensity(self, value: scalar):
        """
        :param value: The new intensity of the laser.
        :returns: None.
        """
        dll.laser_set_intensity(self.instance, c_double(value))

    @property
    def polarization(self):
        """
        :returns: The polarization of the laser.
        """
        return self._polarization

    @polarization.setter
    def polarization(self, value: Polarization):
        """
        :param value: The new polarization of the laser.
        :return: None.
        """
        self._polarization = value
        dll.laser_set_polarization(self.instance, value.instance)

    @property
    def k(self):
        """
        :returns: The direction of the laser. The default direction is ( 1, 0, 0 ).
        """
        return dll.laser_get_k(self.instance)

    @k.setter
    def k(self, value: array_like):
        """
        :param value: The new direction of the laser.
        :return: None.
        """
        _value = np.asarray(value, dtype=float)
        if _value.size != 3:
            raise ValueError('Interaction.k must be a 3d-vector, but has shape {}.'.format(_value.shape))
        dll.laser_set_k(self.instance, _value.flatten())


def _process_hyper_const(hyper_const: array_like) -> ndarray:
    """
    Preprocess the hyperfine-structure constants.

    :param hyper_const: The hyperfine-structure constants. Currently, constants up to the electric quadrupole order are
     supported (A, B). If 'hyper_const' is a scalar,
     it is assumed to be the constant A and the other orders are 0 (MHz).
    :returns: The hyperfine-structure constants as a 3d-vector.
    """
    if hyper_const is None or not hyper_const:
        hyper_const = [0., 0., 0.]
    elif not np.asarray(hyper_const, dtype=float).shape:
        hyper_const = [float(hyper_const), 0., 0.]
    hyper_const = list(hyper_const)
    while len(hyper_const) < 3:
        hyper_const.append(0.)
    return np.asarray(hyper_const, dtype=float)[:3]


# noinspection PyPep8Naming
class Environment:
    def __init__(self, E: array_like = None, B: array_like = None, instance=None):
        """
        Class representing an electromagnetic environment.

        :param E: A static electric field (not implemented).
        :param B: A static magnetic field (linear Zeeman effect).
        :param instance: A pointer to an existing Environment instance.
         If this is specified, the other parameters are omitted.
        """
        self.instance = instance
        if self.instance is None:
            self.instance = dll.environment_construct()
            self.E = E
            self.B = B

    def __del__(self):
        dll.environment_destruct(self.instance)

    @property
    def E(self):
        return dll.environment_get_E(self.instance) * dll.environment_get_e_E(self.instance)
    
    @E.setter
    def E(self, value: array_like):
        if value is None:
            dll.environment_set_E(self.instance, np.asarray([1, 0, 0], dtype=float))
            dll.environment_set_E_double(self.instance, c_double(0.))
        else:
            value = np.asarray(value, dtype=float)
            if not value.shape:
                dll.environment_set_E_double(self.instance, value)
            elif value.shape == (3, ):
                dll.environment_set_E(self.instance, value)
            else:
                raise ValueError('E must be a scalar, 3d-vector or None, but has shape {}'.format(value.shape))

    @property
    def B(self):
        return dll.environment_get_B(self.instance) * dll.environment_get_e_B(self.instance)
    
    @B.setter
    def B(self, value: array_like):
        if value is None:
            dll.environment_set_B(self.instance, np.array([0, 0, 1], dtype=float))
            dll.environment_set_B_double(self.instance, c_double(0.))
        else:
            value = np.asarray(value, dtype=float)
            if not value.shape:
                dll.environment_set_B_double(self.instance, value)
            elif value.shape == (3, ):
                dll.environment_set_B(self.instance, value)
            else:
                raise ValueError('B must be a scalar, 3d-vector or None, but has shape {}'.format(value.shape))


class State:
    def __init__(self, freq_j: scalar, s: scalar, l: scalar, j: scalar, i: scalar, f: scalar, m: scalar,
                 hyper_const: array_like = None, g: scalar = 0, label: str = None, instance=None):
        r"""
        Class representing an atomic quantum state $|(\mathrm{label})SLJIFm\rangle$.

        :param freq_j: The energetic position of the state without the hyperfine structure or the environment (MHz).
        :param s: The electron spin quantum number S.
        :param l: The electronic angular momentum quantum number L.
        :param j: The electronic total angular momentum quantum number J.
        :param i: The nuclear spin quantum number I.
        :param f: The total angular momentum quantum number F.
        :param m: The B-field-axis component quantum number m of the total angular momentum.
        :param hyper_const: The hyperfine-structure constants. Currently, constants up to the electric quadrupole order
         are supported (A, B). If 'hyper_const' is a scalar, it is assumed to be the constant A
         and the other orders are 0 (MHz).
        :param g: The nuclear g-factor.
        :param label: The label of the state. The label is used to link states via decay maps.
        :param instance: A pointer to an existing State instance.
         If this is specified, the other parameters are omitted.
        """
        self.instance = instance
        if self.instance is None:
            tools.check_half_integer(s, l, j, i, f, m)
            hyper_const = _process_hyper_const(hyper_const)
            if label is None:
                label = '{}({}, {}, {})'.format(int(np.around(freq_j, decimals=0)), s, l, j)
            self.instance = dll.state_construct()
            dll.state_init(self.instance, c_double(freq_j), c_double(s), c_double(l), c_double(j), c_double(i),
                           c_double(f), c_double(m), hyper_const, c_double(g), c_char_p(bytes(label, 'utf-8')))

    def __del__(self):
        dll.state_destruct(self.instance)

    def __repr__(self):
        return '{}({})'.format(self.label, ('{}, ' * 6)[:-2]) \
            .format(*[tools.half_integer_to_str(qn, '/') for qn in [self.s, self.l, self.j, self.i, self.f, self.m]])

    def update(self, environment: Environment = None):
        """
        Update the shifted frequency of the state.

        :returns: None.
        """
        if environment is None:
            dll.state_update(self.instance)
        else:
            dll.state_update_env(self.instance, environment.instance)

    def get_shift(self):
        """
        :returns: The difference between the shifted frequency of the hyperfine-structure
         and the frequency of the fine-structure state.
        """
        return dll.state_get_shift(self.instance)

    @property
    def freq_j(self):
        """
        :returns: The frequency of the fine-structure state.
        """
        return dll.state_get_freq_j(self.instance)

    @freq_j.setter
    def freq_j(self, value: scalar):
        dll.state_set_freq_j(self.instance, c_double(value))

    @property
    def freq(self):
        """
        :returns: The shifted frequency of the hyperfine-structure state.
        """
        return dll.state_get_freq(self.instance)

    @property
    def s(self):
        """
        :returns: The electron spin quantum number S.
        """
        return dll.state_get_s(self.instance)

    @property
    def l(self):
        """
        :returns: The electronic angular momentum quantum number L.
        """
        return dll.state_get_l(self.instance)

    @property
    def j(self):
        """
        :returns: The electronic total angular momentum quantum number J.
        """
        return dll.state_get_j(self.instance)

    @property
    def i(self):
        """
        :returns: The nuclear spin quantum number I.
        """
        return dll.state_get_i(self.instance)

    @property
    def f(self):
        """
        :returns: The total angular momentum quantum number F.
        """
        return dll.state_get_f(self.instance)

    @property
    def m(self):
        """
        :returns: The B-field-axis component quantum number m of the total angular momentum.
        """
        return dll.state_get_m(self.instance)

    @property
    def hyper_const(self):
        """
        :returns: The hyperfine-structure constants as a 3d-vector.
        """
        return dll.state_get_hyper_const(self.instance)

    @hyper_const.setter
    def hyper_const(self, value: array_like):
        """
        :param value: The new hyperfine-structure constants. Currently, constants up to the electric quadrupole order
         are supported (A, B). If 'hyper_const' is a scalar, it is assumed to be the constant A
         and the other orders are 0 (MHz).
        :returns: None.
        """
        value = _process_hyper_const(value)
        dll.state_get_hyper_const(self.instance, value)

    @property
    def g(self):
        """
        :returns: The nuclear g-factor.
        """
        return dll.state_get_g(self.instance)

    @g.setter
    def g(self, value: scalar):
        """
        :param value: The new nuclear g-factor.
        :returns: None.
        """
        dll.state_set_g(self.instance, c_double(value))

    @property
    def label(self):
        """
        :returns: The label of the state. The label is used to link states via decay maps.
        """
        return dll.state_get_label(self.instance).decode('utf-8')

    @label.setter
    def label(self, value: str):
        """
        :param value: The label of the state. The label is used to link states via decay maps.
        :returns: None.
        """
        dll.state_set_label(self.instance, c_char_p(bytes(value, 'utf-8')))


def construct_electronic_state(freq_0: scalar, s: scalar, l: scalar, j: scalar, i: scalar = 0,
                               hyper_const: Iterable[scalar] = None, g: scalar = 0, label: str = None):
    """
    Creates all substates of a fine-structure state using a common label.

    :param freq_0: The energetic position of the state without the hyperfine structure or the magnetic field (MHz).
    :param s: The electron spin quantum number S.
    :param l: The electronic angular momentum quantum number L.
    :param j: The electronic total angular momentum quantum number J.
    :param i: The nuclear spin quantum number I.
    :param hyper_const: The hyperfine-structure constants. Currently, constants up to the electric quadrupole order are
     supported (A, B). If 'hyper_const' is a scalar,
     it is assumed to be the constant A and the other orders are 0 (MHz).
    :param g: The nuclear g-factor.
    :param label: The label of the states. The labels are used to link states via decay maps.
    :returns: A list of the created states.
    """
    f = get_f(i, j)
    m = [get_m(_f) for _f in f]
    fm = [(_f, _m) for _f, m_f in zip(f, m) for _m in m_f]
    return [State(freq_0, s, l, j, i, _f, _m, hyper_const=hyper_const, g=g, label=label) for (_f, _m) in fm]


def construct_hyperfine_state(freq_0: scalar, s: scalar, l: scalar, j: scalar, i: scalar, f: scalar,
                              hyper_const: Iterable[scalar] = None, g: scalar = 0, label: str = None):
    """
    Creates all substates of a fine-structure state using a common label.

    :param freq_0: The energetic position of the state without the hyperfine structure or the magnetic field (MHz).
    :param s: The electron spin quantum number S.
    :param l: The electronic angular momentum quantum number L.
    :param j: The electronic total angular momentum quantum number J.
    :param i: The nuclear spin quantum number I.
    :param f: The hyperfine structure total angular momentum quantum number F.
    :param hyper_const: The hyperfine-structure constants. Currently, constants up to the electric quadrupole order are
     supported (A, B). If 'hyper_const' is a scalar,
     it is assumed to be the constant A and the other orders are 0 (MHz).
    :param g: The nuclear g-factor.
    :param label: The label of the states. The labels are used to link states via decay maps.
    :returns: A list of the created states.
    """
    return [State(freq_0, s, l, j, i, f, _m, hyper_const=hyper_const, g=g, label=label) for _m in get_m(f)]


class DecayMap:
    def __init__(self, labels: Iterable[tuple] = None, a: Iterable[scalar] = None, instance=None):
        """
        Class linking sets of atomic states via Einstein-A coefficients.

        :param labels: An iterable of label pairs, corresponding to atomic states which get connected.
        :param a: An Iterable of Einstein-A coefficients (MHz).
        :param instance: A pointer to an existing DecayMap instance.
         If this is specified, the other parameters are omitted.
        """
        self.instance = instance

        if self.instance is None:
            self.instance = dll.decaymap_construct()
            if labels is None:
                labels = []
            if a is None:
                a = []
            self._labels = list(labels)
            for (s0, s1), _a in zip(self._labels, a):
                dll.decaymap_add_decay(self.instance, c_char_p(bytes(s0, 'utf-8')), c_char_p(bytes(s1, 'utf-8')),
                                       c_double(float(_a)))
        else:
            self._labels = self._get_labels()

    def __del__(self):
        dll.decaymap_destruct(self.instance)

    def _get_labels(self):
        """
        :returns: The labels used in the C++ class.
        """
        return [(dll.decaymap_get_label(self.instance, 0, i).decode('utf-8'),
                 dll.decaymap_get_label(self.instance, 1, i).decode('utf-8')) for i in range(self.size)]

    @property
    def labels(self):
        """
        :returns: The list of label pairs, corresponding to atomic states which get connected.
        """
        return self._labels

    @property
    def a(self):
        """
        :returns: The list of Einstein-A coefficients.
        """
        vector_d_p = np.ctypeslib.ndpointer(dtype=float, shape=(self.size, ))
        set_restype(dll.decaymap_get_a, vector_d_p)
        return dll.decaymap_get_a(self.instance).tolist()

    @property
    def size(self):
        """
        :returns: The number of linked sets of atomic states.
        """
        return dll.decaymap_get_size(self.instance)

    def get(self, label_0: str, label_1: str):
        """
        :returns: The number of linked sets of atomic states.
        """
        return dll.decaymap_get_item(
            self.instance, c_char_p(bytes(label_0, 'utf-8')), c_char_p(bytes(label_1, 'utf-8')))


def _gen_label_map(atom):
    """
    :param atom: The atom.
    :returns: A dictionary with state labels as keys
     and an array of the indices of the states with the labels as values.
    """
    if isinstance(atom, int):
        return {'States 0 - {}'.format(atom): np.arange(atom, dtype=int)}
    all_labels = [s.label for s in atom]
    labels = []
    for s in all_labels:
        if s not in labels:
            labels.append(s)
    label_map = {s: np.array([i for i, _s in enumerate(all_labels) if _s == s]) for s in labels}
    return label_map


class Atom:
    def __init__(self, states: Iterable[State] = None, decay_map: DecayMap = None, mass: scalar = 0, instance=None):
        """
        Class representing an Atom and its inner structure.

        :param states: The states of the atom.
        :param decay_map: The decay map which connects the atomic states.
        :param mass: The mass of the atom (u).
        :param instance: A pointer to an existing Atom instance.
         If this is specified, the other parameters are omitted.
        """
        self.instance = instance

        if self.instance is None:
            self.instance = dll.atom_construct()
            if states is None:
                states = []
            if decay_map is None:
                decay_map = DecayMap()
            self.states = list(states)
            self.decay_map = decay_map
            self.mass = mass
            self.label_map = None
            self.update()
        else:
            self._states = None

    def __del__(self):
        dll.atom_destruct(self.instance)

    def __iter__(self):
        for state in self.states:
            yield state

    def __getitem__(self, key: int):
        return self.states[key]

    def update(self):
        """
        Update the atom.

        :returns: None.
        """
        dll.atom_update(self.instance)
        self.label_map = _gen_label_map(self)

    @property
    def states(self):
        """
        :returns: The states of the atom.
        """
        return self._states

    @states.setter
    def states(self, value: Iterable[State]):
        """
        :param value: The new states of the atom.
        :returns: None.
        """
        dll.atom_clear_states(self.instance)
        self._states = list(value)
        for s in self._states:
            dll.atom_add_state(self.instance, s.instance)

    @property
    def decay_map(self):
        """
        :returns: The decay map which connects the atomic states.
        """
        return self._decay_map

    @decay_map.setter
    def decay_map(self, value: DecayMap):
        """
        :param value: The new decay map which connects the atomic states.
        :returns: None.
        """
        self._decay_map = value
        if self._decay_map is None:
            self._decay_map = DecayMap()
        dll.atom_set_decay_map(self.instance, value.instance)

    @property
    def mass(self):
        """
        :returns: The mass of the atom (u).
        """
        return dll.atom_get_mass(self.instance)

    @mass.setter
    def mass(self, value: scalar):
        """
        :param value: The new mass of the atom (u).
        :returns: None.
        """
        dll.atom_set_mass(self.instance, c_double(value))

    @property
    def size(self):
        """
        :returns: The number of states of the atom.
        """
        return dll.atom_get_size(self.instance)

    @property
    def gs(self) -> ndarray:
        """
        :returns: The indices of the ground states.
        """
        vector_i_p = np.ctypeslib.ndpointer(dtype=c_size_t, shape=(dll.atom_get_gs_size(self.instance), ))
        set_restype(dll.atom_get_gs, vector_i_p)
        return dll.atom_get_gs(self.instance)

    @property
    def dipoles(self):
        """
        :returns: The dipole strengths between the atomic states in the 3 basis-components
         of the spherical vector basis ( sigma-, pi, sigma+ ). This can be used to calculate Rabi-frequencies by
         multiplying it with the square-root of a laser intensity in the corresponding polarization.
         The resulting array has shape (3, size, size).
        """
        return np.array([np.ctypeslib.as_array(dll.atom_get_m_dipole(self.instance, c_size_t(i)),
                                               (self.size, self.size)).T for i in range(3)])

    @property
    def l0(self):
        a = dll.atom_get_L0(self.instance)
        return np.ctypeslib.as_array(a, (self.size, self.size)).T

    @property
    def l1(self):
        a = dll.atom_get_L1(self.instance)
        return np.ctypeslib.as_array(a, (self.size, self.size)).T

    def get_y0(self, ground_state_labels: Union[Iterable[str], str] = None) -> np.ndarray:
        """
        :param ground_state_labels: An Iterable of labels belonging to ground states.
        :returns: The initial population of the atom.
        """
        if ground_state_labels is None:
            ground_state_labels = [self.states[0].label]
        indices = np.array([i for i, state in enumerate(self) if state.label in ground_state_labels])
        y0 = np.zeros(self.size)
        y0[indices] = 1 / indices.size
        return y0
    
    def get_y0_mc(self, n_samples: int = None, ground_state_labels: Union[Iterable[str], str] = None) -> np.ndarray:
        """
        :param n_samples: The number of samples to create.
        :param ground_state_labels: An Iterable of labels belonging to ground states.
        :returns: The initial population of the atom for the Monte-Carlo master equation solver.
        """
        if ground_state_labels is None:
            ground_state_labels = [self.states[0].label]
        indices = np.array([i for i, state in enumerate(self) if state.label in ground_state_labels])
        y0 = np.zeros((n_samples, self.size), dtype=complex)
        if n_samples % indices.size:
            raise ValueError(f'The number of samples ({n_samples}) must be a multiple of'
                             f' the number of ground states ({indices.size})')
        batch = int(n_samples / indices.size)
        for i, index in enumerate(indices):
            y0[i * batch:(i + 1) * batch, index] \
                = np.exp(np.random.random(size=batch) * 2 * np.pi * 1j)
        return y0

    def get_state_indexes(self, labels: Union[Iterable[str], str] = None,
                          f: Union[Iterable[scalar], scalar] = None) -> np.ndarray:
        """
        :param labels: The labels of the states whose indexes are to be returned.
        :param f: The F quantum numbers whose indexes are to be returned.
        :returns: The indexes corresponding to the specified labels and F quantum numbers.
        """
        if labels is None:
            labels = set(s.label for s in self.states)
        if f is None:
            f = set(s.f for s in self.states)
        try:
            f = set(f)
        except TypeError:
            f = {f}
        return np.array([i for i, s in enumerate(self.states) if s.label in labels and s.f in f], dtype=int)

    def scattering_rate(self, rho: array_like, theta: array_like = None, phi: array_like = None,
                        as_density_matrix: bool = True, i: array_like = None, j: array_like = None, axis: int = 1):
        """
        Scattering rate of the atom into the direction

            e_r = (sin(theta), cos(theta) * sin(phi), cos(theta) * cos(phi))

        where the z-axis is the quantization axis, which is either (0, 0, 1) or the B-field axis.

        :param rho: The density matrix of the atom. Must have the same size as the atom
         along the specified 'axis' and 'axis' + 1.
        :param theta: The elevation angle of detection relative to the quantization axis.
        :param phi: The azimuthal angle of detection relative to the quantization axis.
        :param as_density_matrix: Whether 'rho' is a state vector or a density matrix.
        :param i: The initially excited state indexes to consider for spontaneous decay.
         If None, all states are considered.
        :param j: The final decayed state indexes to consider for spontaneous decay. If None, all states are considered.
        :param axis: The axis along which the population is aligned in 'rho'.
        :returns: The scattering rate of the atom given the population 'rho' (MHz or Events / s).
        :raises ValueError: 'rho' must have the same size as the atom along the specified 'axis'.
        """
        rho = np.asarray(rho)

        if i is None:
            i = np.arange(self.size, dtype=int)
        else:
            i = np.array(i).flatten()
        if j is None:
            j = np.arange(self.size, dtype=int)
        else:
            j = np.array(j).flatten()

        if axis < 0:
            axis += len(rho.shape)

        l0 = np.array([[1. if _i in i and _j in j else 0. for _i in range(self.size)] for _j in range(self.size)])
        l0 *= self.l0

        if theta is None and phi is None:
            if as_density_matrix:
                rho = np.diagonal(rho.real, axis1=axis, axis2=axis + 1)
                if len(rho.shape) > axis + 1:
                    axes = list(range(len(rho.shape)))
                    axes[axis + 1:] = axes[axis:-1]
                    axes[axis] = len(rho.shape) - 1
                    rho = np.transpose(rho, axes=axes)
            axes = [ax for ax in range(axis)]
            if axes:
                l0 = np.expand_dims(l0, axis=axes)
            axes = [axis + ax + 2 for ax in range(len(rho.shape) - axis - 1)]
            if axes:
                l0 = np.expand_dims(l0, axis=axes)

            sr = tools.transform(l0, rho, axis=axis)
            return np.sum(sr, axis=axis)

        elif theta is None or phi is None:
            raise ValueError('\'theta\' and \'phi\' must either both be specified or both be None.')

        if not as_density_matrix:
            rho = tools.vector_to_diag_matrix(rho, axis=axis)

        a_cart = [[al.a_dipole_cart(self.states[_j].i, self.states[_j].j, self.states[_j].f, self.states[_j].m,
                                    self.states[_i].j, self.states[_i].f, self.states[_i].m)
                   * np.sqrt(2 * self.states[_j].i + 1) * np.sqrt(2 * self.states[_j].j + 1)
                   * np.sqrt(self.decay_map.get(self.states[_j].label, self.states[_i].label))
                   if l0[_j, _i] else np.zeros(3, dtype=complex)
                   if _i in i and _j in j else 0. for _i in range(self.size)] for _j in range(self.size)]
        e_theta = tools.e_theta(theta, phi)
        e_phi = tools.e_phi(theta, phi)

        c_theta = np.array([[np.sum(e_theta * _a_cart) for _a_cart in a_cart_list]
                            for a_cart_list in a_cart])
        c_phi = np.array([[np.sum(e_phi * _a_cart) for _a_cart in a_cart_list]
                          for a_cart_list in a_cart])

        ct_theta = np.array([[np.sum(e_theta * np.conj(_a_cart))
                              for _a_cart in a_cart_list] for a_cart_list in a_cart])
        ct_phi = np.array([[np.sum(e_phi * np.conj(_a_cart))
                            for _a_cart in a_cart_list] for a_cart_list in a_cart])

        axes = [ax for ax in range(axis)]
        if axes:
            c_theta = np.expand_dims(c_theta, axis=axes)
            c_phi = np.expand_dims(c_phi, axis=axes)
            ct_theta = np.expand_dims(ct_theta, axis=axes)
            ct_phi = np.expand_dims(ct_phi, axis=axes)
        axes = [axis + ax + 2 for ax in range(len(rho.shape) - axis - 2)]
        if axes:
            c_theta = np.expand_dims(c_theta, axis=axes)
            c_phi = np.expand_dims(c_phi, axis=axes)
            ct_theta = np.expand_dims(ct_theta, axis=axes)
            ct_phi = np.expand_dims(ct_phi, axis=axes)

        sr = (np.sum([np.expand_dims(tools.get_subarray(c_theta, k, axis), axis=axis + 1)
                      * np.expand_dims(tools.get_subarray(rho, k, axis + 1), axis=axis)
                      for k in range(self.size)], axis=0)
              * np.sum([np.expand_dims(tools.get_subarray(ct_theta, k, axis + 1), axis=axis)
                        * np.expand_dims(tools.get_subarray(rho, k, axis), axis=axis + 1)
                        for k in range(self.size)], axis=0))
        
        sr += (np.sum([np.expand_dims(tools.get_subarray(c_phi, k, axis), axis=axis + 1)
                       * np.expand_dims(tools.get_subarray(rho, k, axis + 1), axis=axis)
                       for k in range(self.size)], axis=0)
               * np.sum([np.expand_dims(tools.get_subarray(ct_phi, k, axis + 1), axis=axis)
                         * np.expand_dims(tools.get_subarray(rho, k, axis), axis=axis + 1)
                         for k in range(self.size)], axis=0))
        return 3 / (8 * np.pi) * np.sum(np.sum(sr, axis=axis), axis=axis).real

    def plot(self, indices: array_like = None, draw_bounds: bool = False, show: bool = True):
        """
        Plot a term scheme of the atom.

        :param indices: The indices of the states to be drawn. If None, all states are drawn.
        :param draw_bounds: Whether to draw the upper vertical bounds of the states.
        :param show: Whether to show the plot.
        :returns: The x and y positions of the states as well as the distance constant d.
        """
        if indices is None:
            indices = np.argsort([state.freq for state in self.states])
        d = 2
        y_i = 0
        y_dict = {}
        m_max = max(s.m for s in self.states)
        ret_x = {}
        ret_y = {}
        for i in indices:
            s = self.states[i]
            key = (s.label, s.s, s.l, s.j, s.i, s.f)
            if key not in y_dict.keys():
                if key[:-1] not in [key_1[:-1] for key_1 in y_dict.keys()]:
                    y_i += 3
                y_dict[key] = y_i * d
                y_i += 1
                if draw_bounds:
                    plt.hlines([y_dict[key] + d * 0.05, y_dict[key] + d * 0.95][1:],
                               xmin=-m_max - 0.5, xmax=m_max + 0.5, ls='--', colors='grey')
            x = np.array([s.m - 0.45, s.m + 0.45])  # + x_off[key[:-1]]
            y = np.array([y_dict[key], y_dict[key]])
            ret_x[i] = x
            ret_y[i] = y
            plt.plot(x, y, 'k-')

        x_ticks = np.linspace(-m_max, m_max, int(2 * m_max + 1), dtype=float)
        plt.xticks(x_ticks)
        y_ticks = np.array([[_y[1], (_y[0][0], _y[0][-1])] for _y in y_dict.items()], dtype=object)
        order = np.argsort(y_ticks, axis=0)[:, 0]
        y_ticks = y_ticks[order]
        plt.yticks(y_ticks[:, 0].astype(float), [str(y_l) for y_l in y_ticks[:, 1]])
        plt.xlabel(r'$m$')
        plt.ylabel(r'$(\mathrm{label}, F)$')
        if show:
            plt.tight_layout()
            plt.show()
        return ret_x, ret_y, d


def _cast_t(t) -> (ndarray, int):
    t = np.asarray(t, dtype=float).flatten()
    t.sort()
    t_size = t.size
    if t[0] != 0:
        if t[0] < 0:
            raise ValueError('All times \'t\' must be positive.')
        t = np.ascontiguousarray(np.concatenate([np.zeros(1, dtype=float), t], axis=0), dtype=float)
    return t, t.size, t_size != t.size


def _cast_delta(delta: array_like, m: Optional[int], size: int) -> ndarray:
    """
    :param delta: An array of frequency shifts for the laser(s). 'delta' must be a scalar or a 1d- or 2d-array
     with shapes (., ) or (., #lasers), respectively.
    :param m: The index of the shifted laser. If delta is a 2d-array, 'm' ist omitted.
    :param size: The number of available lasers.
    :returns: An array of vectors with size 'size' containing frequency shifts for the lasers.
    """
    if delta is None:
        return np.zeros((1, size))
    delta = np.array(delta, dtype=float, order='C')
    if len(delta.shape) != 2 and not -size <= m < size:
        raise IndexError('Laser index \'m\' is out of bounds. Must be {} <= m < {} or None but is {}.'
                         .format(-size, size, m))
    error = False
    if len(delta.shape) > 2:
        error = True
    elif len(delta.shape) == 0:
        if m is None:
            delta = np.full((1, size), delta, dtype=float)
        else:
            delta = np.expand_dims(tools.unit_vector(m, size, dtype=float) * delta, axis=0)
    elif len(delta.shape) == 1:
        if m is None:
            delta = delta[:, None] + np.expand_dims(np.zeros(size), axis=0)
        else:
            delta = tools.unit_vector(np.full(delta.size, m, dtype=int), size, dtype=float) \
                    * np.expand_dims(delta, axis=1)
    elif delta.shape[1] != size:
        error = True
    if error:
        raise ValueError('\'delta\' must be a scalar or a 1d- or 2d-array with shapes '
                         '(., ) or (., #lasers), respectively.')
    return delta


def _cast_y0(y0: Optional[array_like], i_solver: int, atom: Atom):
    """
    :param y0: The initial states of an ensemble of n atoms. Depending on the solver, this must have shape
     i_solver = 0 and 1: (#states, ) or (n, #states).
     i_solver = 2: (#states, ), (n, #states) or (n, #states, #states).
    :param i_solver: The index of the solver.
    :param atom: The atom.
    :returns: The correctly shaped 'y0' for the chosen solver and its C type.
    """
    size = atom.size
    gs = atom.gs

    if i_solver == 0:  # Rate equations.
        if y0 is None:
            y0 = np.zeros(size, dtype=float)
            y0[gs] = 1 / gs.size
            return y0, c_double_p

        y0 = np.array(y0, dtype=float, order='C')

        if not y0.shape or y0.shape[-1] != size:
            raise ValueError('\'y0\' must have size {} in the last axis but has shape {}.'.format(size, y0.shape))

        if len(y0.shape) < 2:  # Add the missing sample axis.
            y0 = np.expand_dims(y0, axis=0)
        elif len(y0.shape) > 2:  # Flatten all but the last axis.
            shape = (sum(y0.shape[:-1]), y0.shape[-1])
            y0 = y0.reshape(shape, order='C')
        y0 /= np.sum(y0, axis=-1)[:, None]
        dtype, pointer = float, c_double_p

    elif i_solver == 1:  # Schroedinger equation / MC master.
        if y0 is None:
            y0 = np.zeros(size, dtype=complex)
            y0[gs[0]] = 1
            return y0, c_complex_p

        y0 = np.array(y0, dtype=complex, order='C')

        if not y0.shape or y0.shape[-1] != size:
            raise ValueError('\'y0\' must have size {} in the last axis but has shape {}.'.format(size, y0.shape))

        if len(y0.shape) < 2:  # Add the missing sample axis.
            y0 = np.expand_dims(y0, axis=0)
        elif len(y0.shape) > 2:  # Flatten all but the last axis.
            shape = (sum(y0.shape[:-1]), y0.shape[-1])
            y0 = y0.reshape(shape, order='C')
        y0 /= tools.absolute_complex(y0, axis=-1)[:, None]  # Normalize
        dtype, pointer = complex, c_complex_p

    elif i_solver == 2:  # Master equation.
        if y0 is None:
            y0 = np.zeros(size, dtype=complex)
            y0[gs] = 1 / gs.size
            return np.diag(y0), c_complex_p

        y0 = np.array(y0, dtype=complex, order='C')

        if not y0.shape or (len(y0.shape) <= 2 and y0.shape[-1] != size) \
                or (len(y0.shape) > 2 and y0.shape[-2:] != (size, size)):
            raise ValueError('\'y0\' must have size {} in the last axis if len(y0.shape) <= 2,'
                             ' or shape {} in the last two axes if len(y0.shape) > 2.'
                             .format(size, (size, size), y0.shape))

        if len(y0.shape) == 2:  # Normalize and create n diagonal matrices.
            y0 /= np.sum(y0, axis=1)[:, None]
            y0 = np.array([np.diag(_y0) for _y0 in y0])
        elif len(y0.shape) > 2:  # Flatten all but the last two axes and normalize.
            shape = (sum(y0.shape[:-2]), y0.shape[-2], y0.shape[-1])
            y0 = y0.reshape(shape, order='C')
            y0 /= np.sum(np.diagonal(y0, axis1=-2, axis2=-1), axis=-1)[:, None, None]
        else:  # Normalize and add the missing sample axis.
            y0 = np.expand_dims(np.diag(y0 / np.sum(y0)), axis=0)
        dtype, pointer = complex, c_complex_p

    else:
        raise ValueError('Solver ' + str(i_solver) + ' not available \'i_solver\' must be in {0, 1, 2}.')

    return np.ascontiguousarray(y0, dtype=dtype), pointer


def _cast_v(v: Optional[array_like]):
    """
    :param v: Atom velocities. Must be a scalar or have shape (n, ) or (n, 3). In the first two cases,
     the velocity vector(s) is assumed to be aligned with the x-axis.
    :returns: The correctly shaped velocities with shape (n, 3).
    :raises ValueError: If 'v' has the wrong shape.
    """
    if v is None:
        return np.array([[0, 0, 0]], dtype=float)
    v = np.array(v, dtype=float, order='C')
    if len(v.shape) == 0:
        return np.array([[v, 0, 0]], dtype=float)
    elif len(v.shape) == 1:
        ret = np.zeros((v.size, 3), dtype=float)
        ret[:, 0] = v
        return ret
    elif len(v.shape) == 2:
        if v.shape[1] == 3:
            return v
    raise ValueError('\'v\' must be a scalar or have shape (n, ) or (n, 3) but has shape {}.'.format(v.shape))
    

class Interaction:
    def __init__(self, atom: Atom = None, lasers: Iterable[Laser] = None, delta_max: scalar = 1e3,
                 controlled: bool = True, instance=None):
        """
        Class representing an Interaction between lasers and an atom.

        :param atom: The atom.
        :param lasers: The lasers.
        :param delta_max: The maximum absolute difference between a laser and a transition frequency
         for that transition to be considered laser-driven (MHz). The default value is 1 GHz.
        :param controlled: Whether the ODE solver uses an error controlled stepper or a fixed step size.
         Setting this to True is particularly useful for dynamics where a changing resolution is required.
         However, this comes at the cost of computing time.
        :param instance: A pointer to an existing Interaction instance.
         If this is specified, the other parameters are omitted.
        """
        self.instance = instance
        if self.instance is None:
            self.instance = dll.interaction_construct()
            self._environment = self._get_environemnt()
            if atom is None:
                atom = Atom()
            if lasers is None:
                lasers = []
            self.atom = atom
            self.lasers = list(lasers)
            self.delta_max = delta_max
            self.controlled = controlled
            self.update()
        else:
            self._environment = self._get_environemnt()
            self._atom = self._get_atom()
            self._lasers = self._get_lasers()
            self.update()

    def __del__(self):
        dll.interaction_destruct(self.instance)

    def _get_environemnt(self):
        """
        :return: The environment used in the C++ class.
        """
        return Environment(instance=dll.interaction_get_environment(self.instance))

    def _get_atom(self):
        """
        :return: The atom used in the C++ class.
        """
        return Atom(instance=dll.interaction_get_atom(self.instance))

    def _get_lasers(self):
        """
        :returns: The lasers used in the C++ class.
        """
        return [Laser(0, instance=dll.interaction_get_laser(self.instance, m))
                for m in range(dll.interaction_get_lasers_size(self.instance))]

    def update(self):
        """
        Updates the Interaction.

        :returns: None.
        """
        dll.interaction_update(self.instance)

    def resonance_info(self):
        r"""
        Prints the detunings of the base frequencies of the lasers in the given atomic system.
        In particular useful for systems with a hyperfine structure. Here $\Delta = \nu_0 - \nu_\mathrm{L}$.

        :returns: None.
        """
        print('Resonance info:')  # \n<label>(S, L, J, I, F, m) -> <label\'>(S\', L\', J\', I\', F\', m\')')
        for k, (laser, laser_m) in enumerate(zip(self.lasers, self.get_rabi())):
            n = 0
            print('Laser {} @ {} MHz:'.format(k, laser.freq))
            for i, state_i in enumerate(self.atom):
                for j, state_j in enumerate(self.atom):
                    if np.abs(laser_m)[i, j] != 0 and i < j:
                        if state_i.freq < state_j.freq:
                            print('{} -> {}: {} MHz'.format(repr(state_i), repr(state_j),
                                                            state_j.freq - state_i.freq - laser.freq))
                        else:
                            print('{} -> {}: {} MHz'.format(repr(state_j), repr(state_i),
                                                            state_i.freq - state_j.freq - laser.freq))
                        n += 1
            if n == 0:
                print('No resonances!')
        print()

    @property
    def environment(self):
        """
        :returns: The environment of the interaction.
        """
        return self._environment

    @environment.setter
    def environment(self, value: Environment):
        """
        :param value: The new environment of the interaction.
        :returns: None.
        """
        self._environment = value
        dll.interaction_set_environment(self.instance, value.instance)

    @property
    def atom(self):
        """
        :returns: The atom of the interaction.
        """
        return self._atom

    @atom.setter
    def atom(self, value: Atom):
        """
        :param value: The new atom of the interaction.
        :returns: None.
        """
        self._atom = value
        dll.interaction_set_atom(self.instance, value.instance)

    @property
    def lasers(self):
        """
        :returns: The lasers of the interaction.
        """
        return self._lasers

    @lasers.setter
    def lasers(self, value: Iterable[Laser]):
        """
        :param value: The new lasers of the interaction.
        :returns: None.
        """
        if value is None:
            value = []
        self._lasers = list(value)
        dll.interaction_clear_lasers(self.instance)
        for laser in self.lasers:
            dll.interaction_add_laser(self.instance, laser.instance)

    @property
    def delta_max(self):
        """
        :returns: The maximum absolute difference between a laser and a transition frequency
         for that transition to be considered laser-driven (MHz). The default value is 1 GHz.
        """
        return dll.interaction_get_delta_max(self.instance)

    @delta_max.setter
    def delta_max(self, value: scalar):
        """
        :param value: The new maximum absolute difference between a laser and a transition frequency
         for that transition to be considered laser-driven (MHz). The default value is 1 GHz.
        :returns: None.
        """
        dll.interaction_set_delta_max(self.instance, c_double(value))

    @property
    def controlled(self):
        """
        :returns: Whether the ODE solver uses an error controlled stepper or a fixed step size.
         Setting this to True is particularly useful for dynamics where a changing resolution is required.
         However, this comes at the cost of computing time.
        """
        return dll.interaction_get_controlled(self.instance)

    @controlled.setter
    def controlled(self, value: bool):
        """
        :param value: Whether the ODE solver uses an error controlled stepper or a fixed step size.
         Setting this to True is particularly useful for dynamics where a changing resolution is required.
         However, this comes at the cost of computing time.
        :returns: None.
        """
        dll.interaction_set_controlled(self.instance, c_bool(value))

    @property
    def dense(self):
        """
        :returns: Whether the ODE solver uses an error controlled dense output stepper.
         If True, this overrides the controlled flag.
         Setting this to True is particularly useful for dynamics where a changing resolution is required.
         However, this comes at the cost of computing time.
        """
        return dll.interaction_get_dense(self.instance)

    @dense.setter
    def dense(self, value: bool):
        """
        :param value: Whether the ODE solver uses an error controlled dense output stepper.
         If True, this overrides the controlled flag.
         Setting this to True is particularly useful for dynamics where a changing resolution is required.
         However, this comes at the cost of computing time.
        :returns: None.
        """
        dll.interaction_set_dense(self.instance, c_bool(value))

    @property
    def dt(self):
        """
        :returns: The (initial) step size of (controlled) solvers.
        """
        return dll.interaction_get_dt(self.instance)

    @dt.setter
    def dt(self, value: scalar):
        """
        :param value: The (initial) step size of (controlled) solvers.
        :returns: None.
        """
        dll.interaction_set_dt(self.instance, c_double(value))

    @property
    def dt_max(self):
        """
        :returns: The maximum step size of controlled solvers.
        """
        return dll.interaction_get_dt_max(self.instance)

    @dt_max.setter
    def dt_max(self, value: scalar):
        """
        :param value: The maximum step size of controlled solvers.
        :returns: None.
        """
        dll.interaction_set_dt_max(self.instance, c_double(value))

    @property
    def atol(self):
        """
        :returns: The absolute error tolerance of controlled solver.
        """
        return dll.interaction_get_atol(self.instance)

    @atol.setter
    def atol(self, value: scalar):
        """
        :param value: The absolute error tolerance of controlled solver.
        :returns: None.
        """
        dll.interaction_set_atol(self.instance, c_double(value))

    @property
    def rtol(self):
        """
        :returns: The relative error tolerance of controlled solver.
        """
        return dll.interaction_get_rtol(self.instance)

    @rtol.setter
    def rtol(self, value: scalar):
        """
        :param value: The relative error tolerance of controlled solver.
        :returns: None.
        """
        dll.interaction_set_rtol(self.instance, c_double(value))

    @property
    def loop(self):
        """
        :returns: Whether there are loops formed by the lasers in the atom.
        """
        return dll.interaction_get_loop(self.instance)

    @property
    def time_dependent(self):
        """
        :returns: Whether the system hamiltonian is allowed to be time dependent.
        """
        return dll.interaction_get_time_dependent(self.instance)

    @time_dependent.setter
    def time_dependent(self, value: bool):
        """
        :param value: Set whether the system hamiltonian is allowed to be time dependent.
        :returns: None.
        """
        dll.interaction_set_time_dependent(self.instance, c_bool(value))

    def get_rabi(self, m: int = None):
        """
        :param m: The laser number 'm'. If None, the Rabi frequencies are returned for all lasers
         as an array with shape (#lasers, atom.size, atom.size).
        :returns: The Rabi frequencies (generated by the laser 'm').
        """
        matrix_cd_p = np.ctypeslib.ndpointer(dtype=np.complex128, shape=(self.atom.size, self.atom.size))
        set_restype(dll.interaction_get_rabi, matrix_cd_p)
        if m is None:
            return np.array([dll.interaction_get_rabi(self.instance, c_size_t(_m))
                             for _m in range(len(self.lasers))], dtype=complex)
        return dll.interaction_get_rabi(self.instance, c_size_t(m))

    @property
    def summap(self):
        """
        :returns: A (atom.size x atom.size)-matrix indicating the states which are laser-connected.
        """
        matrix_i_p = np.ctypeslib.ndpointer(dtype=int, shape=(self.atom.size, self.atom.size))
        set_restype(dll.interaction_get_summap, matrix_i_p)
        return dll.interaction_get_summap(self.instance)

    @property
    def atommap(self):
        """
        :returns: A projection matrix A mapping the state frequencies onto the diagonal of the Hamiltonian.
         It holds diag(H)_i <- sum_j(A_ij * state_j.freq).
        """
        matrix_d_p = np.ctypeslib.ndpointer(dtype=float, shape=(self.atom.size, self.atom.size))
        set_restype(dll.interaction_get_atommap, matrix_d_p)
        return dll.interaction_get_atommap(self.instance).T

    @property
    def deltamap(self):
        """
        :returns: A projection matrix B mapping the laser frequencies onto the diagonal of the Hamiltonian.
         It holds diag(H)_i <- sum_j(B_im * laser_m.freq).
        """
        matrix_d_p = np.ctypeslib.ndpointer(dtype=float, shape=(len(self.lasers), self.atom.size))
        set_restype(dll.interaction_get_deltamap, matrix_d_p)
        return dll.interaction_get_deltamap(self.instance).T

    def get_delta(self):
        vector_d_p = np.ctypeslib.ndpointer(dtype=float, shape=(self.atom.size, ))
        set_restype(dll.interaction_get_delta, vector_d_p)
        return dll.interaction_get_delta(self.instance)

    @property
    def history_size(self):
        """
        :returns: The length of the history of states visited during the generation of the diagonal maps.
        """
        return dll.interaction_get_n_history(self.instance)

    @property
    def history(self):
        """
        :returns: The history of states visited during the generation of the diagonal maps.
        """
        vector_i_p = np.ctypeslib.ndpointer(dtype=c_size_t, shape=(self.history_size, ))
        set_restype(dll.interaction_get_history, vector_i_p)
        return dll.interaction_get_history(self.instance)

    def hamiltonian(self, t: array_like, delta: array_like = None, m: Optional[int] = 0, v: array_like = None):
        """
        :param t: The times when to compute the solution.
        :param delta: An array of frequency shifts for the laser(s). 'delta' must be a scalar or a 1d- or 2d-array
         with shapes (n, ) or (n, #lasers), respectively.
        :param m: The index of the shifted laser. If delta is a 2d-array, 'm' ist omitted.
        :param v: Atom velocities. Must be a scalar or have shape (n, ) or (n, 3). In the first two cases,
         the velocity vector(s) are assumed to be aligned with the x-axis.
        :returns: The integrated master equation as a complex-valued array of shape (n, #states, #states, #times).
        """
        t, t_size, ex = _cast_t(t)

        if isinstance(delta, np.ndarray) and isinstance(v, np.ndarray):
            if delta.shape == v.shape:
                if delta.flags.f_contiguous:
                    delta = np.ascontiguousarray(delta)
                if v.flags.f_contiguous:
                    v = np.ascontiguousarray(v)
            sample_size = delta.shape[0]
        else:
            delta = _cast_delta(delta, m, len(self.lasers))
            v = _cast_v(v)

            sample_size = max([delta.shape[0], v.shape[0], 1])

            delta = np.array(np.broadcast_to(delta, (sample_size, len(self.lasers))), dtype=float, order='C')
            v = np.array(np.broadcast_to(v, (sample_size, 3)), dtype=float, order='C')

        results = np.zeros((sample_size, self.atom.size, self.atom.size, t_size), dtype=complex)
        dll.interaction_get_hamiltonian(self.instance, t.ctypes.data_as(c_double_p), delta.ctypes.data_as(c_double_p),
                                        v.ctypes.data_as(c_double_p), results.ctypes.data_as(c_complex_p),
                                        c_size_t(t_size), c_size_t(sample_size))
        if ex:
            results = results[:, :, :, 1:]
        return results

    def rates(self, t: array_like, delta: array_like = None, m: Optional[int] = 0, v: array_like = None,
              y0: array_like = None):
        """
        Solver for the rate equations. Solutions for n samples can be calculated in parallel.

        :param t: The times when to compute the solution (us).
        :param delta: An array of frequency shifts for the laser(s) (MHz).
         'delta' must be a scalar or a 1d- or 2d-array with shapes (n, ) or (n, #lasers), respectively.
        :param m: The index of the shifted laser. If delta is a 2d-array, 'm' ist omitted.
        :param v: Atom velocities (m/s). Must be a scalar or have shape (n, ) or (n, 3). In the first two cases,
         the velocity vector(s) is(are) assumed to be aligned with the x-axis.
        :param y0: The initial state of the atom. This must be None or have shape (#states, ) or (n, #states).
         If None, the ground states are populated equally.
        :returns: The integrated rate equations as a real-valued array of shape (n, #states, #times).
        """
        t, t_size, ex = _cast_t(t)

        if isinstance(delta, np.ndarray) and isinstance(v, np.ndarray) and isinstance(y0, np.ndarray):
            if delta.shape == v.shape == y0.shape:
                if delta.flags.f_contiguous:
                    delta = np.ascontiguousarray(delta)
                if v.flags.f_contiguous:
                    v = np.ascontiguousarray(v)
                if y0.flags.f_contiguous:
                    y0 = np.ascontiguousarray(y0)
            sample_size = delta.shape[0]
        else:
            delta = _cast_delta(delta, m, len(self.lasers))
            v = _cast_v(v)
            y0, ctype = _cast_y0(y0, 0, self.atom)

            sample_size = max([delta.shape[0], v.shape[0], 1])

            delta = np.array(np.broadcast_to(delta, (sample_size, len(self.lasers))), dtype=float, order='C')
            v = np.array(np.broadcast_to(v, (sample_size, 3)), dtype=float, order='C')
            y0 = np.array(np.broadcast_to(y0, (sample_size, self.atom.size)), dtype=float, order='C')

        results = np.zeros((sample_size, self.atom.size, t_size), dtype=float)
        dll.interaction_rates(self.instance, t.ctypes.data_as(c_double_p), delta.ctypes.data_as(c_double_p),
                              v.ctypes.data_as(c_double_p), y0.ctypes.data_as(c_double_p),
                              results.ctypes.data_as(c_double_p), c_size_t(t_size), c_size_t(sample_size))
        if ex:
            results = results[:, :, 1:]
        return results

    def schroedinger(self, t: array_like, delta: array_like = None, m: Optional[int] = 0, v: array_like = None,
                     y0: array_like = None):
        """
        Solver for the Schroedinger equation. Solutions for n samples can be calculated in parallel.

        :param t: The times when to compute the solution (us).
        :param delta: An array of frequency shifts for the laser(s) (MHz).
         'delta' must be a scalar or a 1d- or 2d-array with shapes (n, ) or (n, #lasers), respectively.
        :param m: The index of the shifted laser. If delta is a 2d-array, 'm' ist omitted.
        :param v: Atom velocities (m/s). Must be a scalar or have shape (n, ) or (n, 3). In the first two cases,
         the velocity vector(s) is(are) assumed to be aligned with the x-axis.
        :param y0: The initial state of the atom. This must be None or have shape (n, #states).
         If None, only the first ground state is populated.
        :returns: The integrated Schroedinger equation as a complex-valued array of shape (n, #states, #times).
        """
        t, t_size, ex = _cast_t(t)

        if isinstance(delta, np.ndarray) and isinstance(v, np.ndarray) and isinstance(y0, np.ndarray):
            if delta.shape == v.shape == y0.shape:
                if delta.flags.f_contiguous:
                    delta = np.ascontiguousarray(delta)
                if v.flags.f_contiguous:
                    v = np.ascontiguousarray(v)
                if y0.flags.f_contiguous:
                    y0 = np.ascontiguousarray(y0)
            sample_size = delta.shape[0]
        else:
            delta = _cast_delta(delta, m, len(self.lasers))
            v = _cast_v(v)
            y0, ctype = _cast_y0(y0, 1, self.atom)

            sample_size = max([delta.shape[0], v.shape[0], y0.shape[0], 1])

            delta = np.array(np.broadcast_to(delta, (sample_size, len(self.lasers))), dtype=float, order='C')
            v = np.array(np.broadcast_to(v, (sample_size, 3)), dtype=float, order='C')
            y0 = np.array(np.broadcast_to(y0, (sample_size, self.atom.size)), dtype=complex, order='C')

        results = np.zeros((sample_size, self.atom.size, t_size), dtype=complex)
        dll.interaction_schroedinger(self.instance, t.ctypes.data_as(c_double_p), delta.ctypes.data_as(c_double_p),
                                     v.ctypes.data_as(c_double_p), y0.ctypes.data_as(c_complex_p),
                                     results.ctypes.data_as(c_complex_p), c_size_t(t_size), c_size_t(sample_size))
        if ex:
            results = results[:, :, 1:]
        return results

    def master(self, t: array_like, delta: array_like = None, m: Optional[int] = 0, v: array_like = None,
               y0: array_like = None):
        """
        Solver for the master equation. Solutions for n samples can be calculated in parallel.

        :param t: The times when to compute the solution (us).
        :param delta: An array of frequency shifts for the laser(s) (MHz).
         'delta' must be a scalar or a 1d- or 2d-array with shapes (n, ) or (n, #lasers), respectively.
        :param m: The index of the shifted laser. If delta is a 2d-array, 'm' ist omitted.
        :param v: Atom velocities (m/s). Must be a scalar or have shape (n, ) or (n, 3). In the first two cases,
         the velocity vector(s) is(are) assumed to be aligned with the x-axis.
        :param y0: The initial state / density matrix of the atom.
         This must be None or have shape (#states, ), (n or 1, #states) or (n or 1, #states, #states).
         If #states == n, it is interpreted as (n, #states). If None, the ground states are populated equally.
        :returns: The integrated master equation as a complex-valued array of shape (n, #states, #states, #times).
        """
        t, t_size, ex = _cast_t(t)

        if isinstance(delta, np.ndarray) and isinstance(v, np.ndarray) and isinstance(y0, np.ndarray):
            if delta.shape == v.shape == y0.shape:
                if delta.flags.f_contiguous:
                    delta = np.ascontiguousarray(delta)
                if v.flags.f_contiguous:
                    v = np.ascontiguousarray(v)
                if y0.flags.f_contiguous:
                    y0 = np.ascontiguousarray(y0)
            sample_size = delta.shape[0]
        else:
            delta = _cast_delta(delta, m, len(self.lasers))
            v = _cast_v(v)
            y0, ctype = _cast_y0(y0, 2, self.atom)

            sample_size = max([delta.shape[0], v.shape[0], 1])

            delta = np.array(np.broadcast_to(delta, (sample_size, len(self.lasers))), dtype=float, order='C')
            v = np.array(np.broadcast_to(v, (sample_size, 3)), dtype=float, order='C')
            y0 = np.array(np.broadcast_to(y0, (sample_size, self.atom.size, self.atom.size)), dtype=complex, order='C')

        results = np.zeros((sample_size, self.atom.size, self.atom.size, t_size), dtype=complex)
        dll.interaction_master(self.instance, t.ctypes.data_as(c_double_p), delta.ctypes.data_as(c_double_p),
                               v.ctypes.data_as(c_double_p), y0.ctypes.data_as(c_complex_p),
                               results.ctypes.data_as(c_complex_p), c_size_t(t_size), c_size_t(sample_size))
        if ex:
            results = results[:, :, :, 1:]
        return results

    def mc_master(self, t: array_like, delta: array_like = None, m: Optional[int] = 0, v: array_like = None,
                  y0: array_like = None, dynamics: bool = False, ntraj: int = 500, as_density_matrix: bool = False):
        """
        Solver for the Monte-Carlo master equation. Solutions for n samples can be calculated in parallel.

        :param t: The times when to compute the solution (us).
        :param delta: An array of frequency shifts for the laser(s) (MHz).
         'delta' must be a scalar or a 1d- or 2d-array with shapes (n, ) or (n, #lasers), respectively.
        :param m: The index of the shifted laser. If delta is a 2d-array, 'm' ist omitted.
        :param v: Atom velocities (m/s). Must be a scalar or have shape (n, ) or (n, 3). In the first two cases,
         the velocity vector(s) is(are) assumed to be aligned with the x-axis.
        :param y0: The initial state of the atom. This must be None or have shape (n, #states).
         If None, only the first ground state is populated.
        :param dynamics: Whether to compute the dynamics of the photon-atom interactions.
        :param ntraj: The number of samples to compute if no samples were given with 'delta', 'v', or 'y0'.
        :param as_density_matrix: Whether the result is returned as density matrices or as state vectors.
        :returns: The integrated MC-Schroedinger equation as a complex-valued array of shape (n, #states, #times).
        """
        if self.controlled:
            raise ValueError('Controlled or Dense steppers are not supported for \'mc_master\'.'
                             ' Decrease the step size if necessary.')
        if dynamics and self.atom.mass <= 0:
            raise ValueError('To simulate mechanical dynamics, the mass of the atom must be specified.')

        t, t_size, ex = _cast_t(t)

        if isinstance(delta, np.ndarray) and isinstance(v, np.ndarray) and isinstance(y0, np.ndarray):
            if delta.shape == v.shape == y0.shape:
                if delta.flags.f_contiguous:
                    delta = np.ascontiguousarray(delta)
                if v.flags.f_contiguous:
                    v = np.ascontiguousarray(v)
                if y0.flags.f_contiguous:
                    y0 = np.ascontiguousarray(y0)
            sample_size = delta.shape[0]
        else:
            delta = _cast_delta(delta, m, len(self.lasers))
            v = _cast_v(v)
            y0, ctype = _cast_y0(y0, 1, self.atom)

            sample_size = max([delta.shape[0], v.shape[0], y0.shape[0], 1])
            if sample_size == 1:
                sample_size = ntraj

            delta = np.array(np.broadcast_to(delta, (sample_size, len(self.lasers))), dtype=float, order='C')
            v = np.array(np.broadcast_to(v, (sample_size, 3)), dtype=float, order='C')
            y0 = np.array(np.broadcast_to(y0, (sample_size, self.atom.size)), dtype=complex, order='C')

        results = np.zeros((sample_size, self.atom.size, t_size), dtype=complex)
        dll.interaction_mc_master(self.instance, t.ctypes.data_as(c_double_p), delta.ctypes.data_as(c_double_p),
                                  v.ctypes.data_as(c_double_p), y0.ctypes.data_as(c_complex_p), c_bool(dynamics),
                                  results.ctypes.data_as(c_complex_p), c_size_t(t_size), c_size_t(sample_size))

        if ex:
            results = results[:, :, 1:]

        if as_density_matrix:
            results = results[:, :, None, :] * results[:, None, :, :].conj()

        return results, v

    def scattering_rate(self, rho: array_like, theta: array_like = None, phi: array_like = None,
                        as_density_matrix: bool = True, i: array_like = None, j: array_like = None, axis: int = 1):
        """
        Scattering rate of the atom into the direction

            e_r = (sin(theta), cos(theta) * sin(phi), cos(theta) * cos(phi))

        where the z-axis is the quantization axis, which is either (0, 0, 1) or the B-field axis.

        :param rho: The density matrix of the atom. Must have the same size as the atom
         along the specified 'axis' and 'axis' + 1.
        :param theta: The elevation angle of detection relative to the quantization axis.
        :param phi: The azimuthal angle of detection relative to the quantization axis.
        :param as_density_matrix: Whether 'rho' is a state vector or a density matrix.
        :param i: The initially excited state indexes to consider for spontaneous decay.
         If None, all states are considered.
        :param j: The final decayed state indexes to consider for spontaneous decay. If None, all states are considered.
        :param axis: The axis along which the population is aligned in 'rho'.
        :returns: The scattering rate of the atom given the population 'rho' (MHz or Events / s).
        :raises ValueError: 'rho' must have the same size as the atom along the specified 'axis'.
        """
        return self.atom.scattering_rate(rho, theta=theta, phi=phi, as_density_matrix=as_density_matrix,
                                         i=i, j=j, axis=axis)


def _define_colors(n: int, label_map: dict, colormap: str = None):
    """
    :param n: The size of the system.
    :param label_map: A dictionary with state labels as keys
     and an array of the indices of the states with the labels as values.
    :param colormap: A matplotlib colormap.
    :returns: A list of colors with size n.
    """
    cmap = cm.get_cmap(colormap)
    labels = [(k, v) for k, v in label_map.items()]
    labels = sorted(labels, key=lambda kv: min(kv[1]))
    colors = ['', ] * n
    for i, (label, indices) in enumerate(labels):
        for index in indices:
            if colormap is None:
                colors[index] = tools.COLORS.PYPLOT[i % 10]
            else:
                colors[index] = cmap(i / (len(labels) - 1))
    return colors



# -*- coding: utf-8 -*-
"""
PyCLS.cpp.cpp
=============

Python/C++ interface.
"""

import os
import ctypes
from numpy import ctypeslib


def set_argtypes(func, argtypes):
    func.argtypes = argtypes


def set_restype(func, restype):
    func.restype = restype


# noinspection PyPep8Naming
class c_complex(ctypes.Structure):
    """
    Complex number, compatible with std::complex layout.
    """
    _fields_ = [('real', ctypes.c_double), ('imag', ctypes.c_double)]

    def __init__(self, z):
        """
        :param z: A complex scalar.
        """
        super().__init__()
        self.real = z.real
        self.imag = z.imag

    def to_complex(self):
        """
        Convert to Python complex.

        :returns: A Python-type complex scalar.
        """
        return self.real + 1.j * self.imag


c_bool = ctypes.c_bool
c_bool_p = ctypes.POINTER(ctypes.c_bool)
c_int = ctypes.c_int
c_int_p = ctypes.POINTER(ctypes.c_int)
c_size_t = ctypes.c_size_t
c_size_t_p = ctypes.POINTER(ctypes.c_size_t)
c_float = ctypes.c_float
c_float_p = ctypes.POINTER(ctypes.c_float)
c_double = ctypes.c_double
c_double_p = ctypes.POINTER(ctypes.c_double)
c_complex_p = ctypes.POINTER(c_complex)
c_char_p = ctypes.c_char_p

vector3d_p = ctypeslib.ndpointer(dtype=float, shape=(3, ))
vector3cd_p = ctypeslib.ndpointer(dtype=complex, shape=(3, ))
matrix3cd_p = ctypeslib.ndpointer(dtype=complex, shape=(3, 3))

PolarizationHandler = ctypes.POINTER(ctypes.c_char)
LaserHandler = ctypes.POINTER(ctypes.c_char)
EnvironmentHandler = ctypes.POINTER(ctypes.c_char)
StateHandler = ctypes.POINTER(ctypes.c_char)
DecayMapHandler = ctypes.POINTER(ctypes.c_char)
AtomHandler = ctypes.POINTER(ctypes.c_char)
InteractionHandler = ctypes.POINTER(ctypes.c_char)
ResultHandler = ctypes.POINTER(ctypes.c_char)
SpectrumHandler = ctypes.POINTER(ctypes.c_char)
MultivariateNormalHandler = ctypes.POINTER(ctypes.c_char)

dll_path = os.path.abspath(os.path.dirname(__file__))
x64 = r'\x64' if ctypes.sizeof(ctypes.c_void_p) == 8 else ''
dll_path = os.path.join(dll_path, r'src\qspec_cpp{}\Release'.format(x64))
dll_name = 'qspec_cpp.dll'
dll = ctypes.CDLL(os.path.join(dll_path, dll_name))


""" simulate """
# Polarization
dll.polarization_construct.restype = PolarizationHandler
dll.polarization_destruct.argtypes = (PolarizationHandler, )

dll.polarization_init.argtypes = (PolarizationHandler, vector3cd_p, vector3d_p, c_bool)
dll.polarization_def_q_axis.argtypes = (PolarizationHandler, vector3d_p, c_bool)

dll.polarization_get_q_axis.argtypes = (PolarizationHandler, )
dll.polarization_get_q_axis.restype = vector3d_p

dll.polarization_get_x.argtypes = (PolarizationHandler, )
dll.polarization_get_x.restype = vector3cd_p

dll.polarization_get_q.argtypes = (PolarizationHandler, )
dll.polarization_get_q.restype = vector3cd_p

# Laser
dll.laser_construct.restype = LaserHandler
dll.laser_destruct.argtypes = (LaserHandler, )

dll.laser_init.argtype = (LaserHandler, c_double, c_double, PolarizationHandler, c_double_p)

dll.laser_get_freq.argtypes = (LaserHandler, )
dll.laser_get_freq.restype = c_double
dll.laser_set_freq.argtypes = (LaserHandler, c_double)

dll.laser_get_intensity.argtypes = (LaserHandler, )
dll.laser_get_intensity.restype = c_double
dll.laser_set_intensity.argtypes = (LaserHandler, c_double)

dll.laser_get_polarization.argtypes = (LaserHandler, )
dll.laser_get_polarization.restype = PolarizationHandler
dll.laser_set_polarization.argtypes = (LaserHandler, PolarizationHandler)

dll.laser_get_k.argtypes = (LaserHandler, )
dll.laser_get_k.restype = vector3d_p
dll.laser_set_k.argtypes = (LaserHandler, vector3d_p)


# Environment
dll.environment_construct.restype = EnvironmentHandler
dll.environment_destruct.argtypes = (EnvironmentHandler, )

dll.environment_get_E.argtypes = (EnvironmentHandler, )
dll.environment_get_E.restype = c_double
dll.environment_get_e_E.argtypes = (EnvironmentHandler, )
dll.environment_get_e_E.restype = vector3d_p
dll.environment_set_E.argtypes = (EnvironmentHandler, vector3d_p)
dll.environment_set_E_double.argtypes = (EnvironmentHandler, c_double)

dll.environment_get_B.argtypes = (EnvironmentHandler, )
dll.environment_get_B.restype = c_double
dll.environment_get_e_B.argtypes = (EnvironmentHandler, )
dll.environment_get_e_B.restype = vector3d_p
dll.environment_set_B.argtypes = (EnvironmentHandler, vector3d_p)
dll.environment_set_B_double.argtypes = (EnvironmentHandler, c_double)


# State
dll.state_construct.restype = StateHandler
dll.state_destruct.argtypes = (StateHandler, )

dll.state_init.argtypes = (StateHandler, c_double, c_double, c_double, c_double, c_double, c_double,
                           c_double, vector3d_p, c_double, c_char_p)

dll.state_update.argtypes = (StateHandler, )
dll.state_update_env.argtypes = (StateHandler, EnvironmentHandler)

dll.state_get_freq_j.argtypes = (StateHandler, )
dll.state_get_freq_j.restype = c_double
dll.state_set_freq_j.argtypes = (StateHandler, c_double)

dll.state_get_freq.argtypes = (StateHandler, )
dll.state_get_freq.restype = c_double

dll.state_get_s.argtypes = (StateHandler, )
dll.state_get_s.restype = c_double

dll.state_get_l.argtypes = (StateHandler, )
dll.state_get_l.restype = c_double

dll.state_get_j.argtypes = (StateHandler, )
dll.state_get_j.restype = c_double

dll.state_get_i.argtypes = (StateHandler, )
dll.state_get_i.restype = c_double

dll.state_get_f.argtypes = (StateHandler, )
dll.state_get_f.restype = c_double

dll.state_get_m.argtypes = (StateHandler, )
dll.state_get_m.restype = c_double

dll.state_get_hyper_const.argtypes = (StateHandler, )
dll.state_get_hyper_const.restype = vector3d_p
dll.state_set_hyper_const.argtypes = (StateHandler, vector3d_p)

dll.state_get_g.argtypes = (StateHandler, )
dll.state_get_g.restype = c_double
dll.state_set_g.argtypes = (StateHandler, c_double)

dll.state_get_label.argtypes = (StateHandler, )
dll.state_get_label.restype = c_char_p
dll.state_set_label.argtypes = (StateHandler, c_char_p)


# DecayMap
dll.decaymap_construct.restype = DecayMapHandler
dll.decaymap_destruct.argtypes = (DecayMapHandler, )

dll.decaymap_add_decay.argtypes = (DecayMapHandler, c_char_p, c_char_p, c_double)

dll.decaymap_get_label.argtypes = (DecayMapHandler, c_size_t, c_size_t)
dll.decaymap_get_label.restype = c_char_p

dll.decaymap_get_a.argtypes = (DecayMapHandler, )

dll.decaymap_get_size.argtypes = (DecayMapHandler, )
dll.decaymap_get_size.restype = c_size_t

dll.decaymap_get_item.argtypes = (DecayMapHandler, c_char_p, c_char_p)
dll.decaymap_get_item.restype = c_double


# Atom
dll.atom_construct.restype = AtomHandler
dll.atom_destruct.argtypes = (AtomHandler, )

dll.atom_update.argtypes = (AtomHandler, )

dll.atom_add_state.argtypes = (AtomHandler, StateHandler)
dll.atom_clear_states.argtypes = (AtomHandler, )

dll.atom_get_decay_map.argtypes = (AtomHandler, )
dll.atom_get_decay_map.restype = DecayMapHandler
dll.atom_set_decay_map.argtypes = (AtomHandler, DecayMapHandler)

dll.atom_get_mass.argtypes = (AtomHandler, )
dll.atom_get_mass.restype = c_double
dll.atom_set_mass.argtypes = (AtomHandler, c_double)

dll.atom_get_size.argtypes = (AtomHandler, )
dll.atom_get_size.restype = c_size_t

dll.atom_get_gs_size.argtypes = (AtomHandler, )
dll.atom_get_gs.restype = c_size_t
dll.atom_get_gs.argtypes = (AtomHandler, )
dll.atom_get_gs.restype = c_size_t_p

dll.atom_get_m_dipole.argtypes = (AtomHandler, c_size_t)
dll.atom_get_m_dipole.restype = c_double_p

dll.atom_get_L0.argtypes = (AtomHandler, )
dll.atom_get_L0.restype = c_double_p

dll.atom_get_L1.argtypes = (AtomHandler, )
dll.atom_get_L1.restype = c_double_p


# Interaction
dll.interaction_construct.restype = InteractionHandler
dll.interaction_destruct.argtypes = (InteractionHandler, )

dll.interaction_update.argtypes = (InteractionHandler, )

dll.interaction_get_environment.argtypes = (InteractionHandler, )
dll.interaction_get_environment.restype = EnvironmentHandler
dll.interaction_set_environment.argtypes = (InteractionHandler, EnvironmentHandler)

dll.interaction_set_atom.argtypes = (InteractionHandler, AtomHandler)

dll.interaction_add_laser.argtypes = (InteractionHandler, LaserHandler)
dll.interaction_clear_lasers.argtypes = (InteractionHandler, )
dll.interaction_get_lasers_size.argtyps = (InteractionHandler, )
dll.interaction_get_lasers_size.restype = c_size_t
dll.interaction_get_laser.argtypes = (InteractionHandler, c_size_t)
dll.interaction_get_laser.restype = LaserHandler

dll.interaction_get_delta_max.argtypes = (InteractionHandler, )
dll.interaction_get_delta_max.restype = c_double
dll.interaction_set_delta_max.argtypes = (InteractionHandler, c_double)

dll.interaction_get_controlled.argtypes = (InteractionHandler, )
dll.interaction_get_controlled.restype = c_bool
dll.interaction_set_controlled.argtypes = (InteractionHandler, c_bool)

dll.interaction_get_dense.argtypes = (InteractionHandler, )
dll.interaction_get_dense.restype = c_bool
dll.interaction_set_dense.argtypes = (InteractionHandler, c_bool)

dll.interaction_get_dt.argtypes = (InteractionHandler, )
dll.interaction_get_dt.restype = c_double
dll.interaction_set_dt.argtypes = (InteractionHandler, c_double)

dll.interaction_get_dt_max.argtypes = (InteractionHandler, )
dll.interaction_get_dt_max.restype = c_double
dll.interaction_set_dt_max.argtypes = (InteractionHandler, c_double)

dll.interaction_get_atol.argtypes = (InteractionHandler, )
dll.interaction_get_atol.restype = c_double
dll.interaction_set_atol.argtypes = (InteractionHandler, c_double)

dll.interaction_get_rtol.argtypes = (InteractionHandler, )
dll.interaction_get_rtol.restype = c_double
dll.interaction_set_rtol.argtypes = (InteractionHandler, c_double)

dll.interaction_get_n_history.argtypes = (InteractionHandler, )
dll.interaction_get_n_history.restype = c_int
dll.interaction_get_history.argtypes = (InteractionHandler, )

dll.interaction_get_loop.argtypes = (InteractionHandler, )
dll.interaction_get_loop.restype = c_bool

dll.interaction_get_time_dependent.argtypes = (InteractionHandler, )
dll.interaction_get_time_dependent.restype = c_bool
dll.interaction_set_time_dependent.argtypes = (InteractionHandler, c_bool)

dll.interaction_get_summap.argtypes = (InteractionHandler, )
dll.interaction_get_rabi.argtypes = (InteractionHandler, c_size_t)
dll.interaction_get_atommap.argtypes = (InteractionHandler, )
dll.interaction_get_deltamap.argtypes = (InteractionHandler, )
dll.interaction_get_delta.argtypes = (InteractionHandler, )

dll.interaction_get_hamiltonian.argtypes = \
    (InteractionHandler, c_double_p, c_double_p, c_double_p, c_complex_p, c_size_t, c_size_t)
dll.interaction_get_hamiltonian.restype = ctypes.c_void_p

dll.interaction_rates.argtypes = \
    (InteractionHandler, c_double_p, c_double_p, c_double_p, c_double_p, c_double_p, c_size_t, c_size_t)
dll.interaction_rates.restype = ctypes.c_void_p

dll.interaction_schroedinger.argtypes = \
    (InteractionHandler, c_double_p, c_double_p, c_double_p, c_complex_p, c_complex_p, c_size_t, c_size_t)
dll.interaction_schroedinger.restype = ctypes.c_void_p

dll.interaction_master.argtypes = \
    (InteractionHandler, c_double_p, c_double_p, c_double_p, c_complex_p, c_complex_p, c_size_t, c_size_t)
dll.interaction_master.restype = ctypes.c_void_p

dll.interaction_mc_master.argtypes = \
    (InteractionHandler, c_double_p, c_double_p, c_double_p, c_complex_p, c_bool, c_complex_p, c_size_t, c_size_t)
dll.interaction_mc_master.restype = ctypes.c_void_p


# ScatteringRate
dll.sr_generate_y.argtypes = (c_complex_p, c_complex_p, c_complex_p, c_size_t_p, c_size_t_p, c_double_p)
dll.sr_generate_y.restype = ctypes.c_void_p


# King
dll.multivariatenormal_construct.argtypes = (c_double_p, c_double_p, c_size_t)
dll.multivariatenormal_construct.restype = MultivariateNormalHandler
dll.multivariatenormal_destruct.argtypes = (MultivariateNormalHandler, )

dll.multivariatenormal_size.restype = c_size_t
dll.multivariatenormal_rvs.restype = ctypes.c_void_p

dll.gen_collinear.argtypes = (c_double_p, c_double_p, c_double_p, c_size_t_p, c_size_t, c_size_t,
                              c_size_t_p, c_bool, c_size_t, c_bool)
dll.gen_collinear.restype = ctypes.c_size_t



# -*- coding: utf-8 -*-
"""
qspec._analyze_cpp
==================

Classes and methods for the 'analyze' module using the Python/C++ interface.
"""

import numpy as np

from qspec.qtypes import *
from qspec._cpp import *

__all__ = ['generate_collinear_points_cpp']


class MultivariateNormal:
    def __init__(self, mean: array_like, cov: array_like, instance=None):
        self.instance = instance
        if self.instance is None:
            mean, cov = np.ascontiguousarray(mean, dtype=float), np.ascontiguousarray(cov, dtype=float)
            self.instance = dll.multivariatenormal_construct(
                mean.ctypes.data_as(c_double_p), cov.ctypes.data_as(c_double_p), c_size_t(mean.size))

    def __del__(self):
        dll.multivariatenormal_destruct(self.instance)

    @property
    def size(self):
        return dll.multivariatenormal_size(self.instance)

    def rvs(self):
        ret = np.zeros(self.size, dtype=float)
        dll.multivariatenormal_rvs(self.instance, ret.ctypes.data_as(c_double_p))
        return ret


def generate_collinear_points_cpp(mean: ndarray, cov: ndarray, n_samples: int = None, n_accepted: int = None,
                                  seed: int = None, report: bool = None, **kwargs):
    """
    :param mean: The data vectors. Must have shape (k, n), where k is the number of data points
     and n is the number of dimensions of each point.
    :param cov: The covariance matrices of the data vectors. Must have shape (k, n, n).
     Use 'covariance_matrix' to construct covariance matrices.
    :param n_samples: The number of samples generated for each data point.
    :param n_accepted: The number of samples to be accepted for each data point.
    :param seed: A seed for the random number generator.
    :param report: Whether to report the number of samples.
    :param kwargs: Additional keyword arguments.
    :returns: The randomly generated data vectors p with shape (n_accepted, k ,n) aligned along a straight line
     and the number of accepted and generated samples.
    """
    mean, cov = np.ascontiguousarray(mean, dtype=float), np.ascontiguousarray(cov, dtype=float)
    size = mean.shape[0]
    dim = mean.shape[1]

    if n_samples is None and n_accepted is None:
        n_samples, n_accepted = 100000, 100000
    elif n_samples is None:
        n_samples = 0
    elif n_accepted is None:
        n_accepted = max(100000, n_samples)

    if seed is None:
        user_seed, seed = False, 0
    else:
        user_seed, seed = True, int(seed)

    n_target = n_accepted
    x = np.zeros((n_accepted, size, dim), dtype=float)
    n_accepted = c_size_t(n_accepted)
    n_samples = c_size_t(n_samples)
    dll.gen_collinear(x.ctypes.data_as(c_double_p), mean.ctypes.data_as(c_double_p), cov.ctypes.data_as(c_double_p),
                      c_size_t_p(n_accepted), c_size_t(size), c_size_t(dim),
                      c_size_t_p(n_samples), c_bool(user_seed), c_size_t(seed), c_bool(report))
    if n_accepted.value < n_target:
        x = x[:n_accepted.value]
    return x, n_accepted.value, n_samples.value



# -*- coding: utf-8 -*-
"""
qspec._simulate
===============

Module for simulations of laser-atom interaction.
"""

from time import time
import numpy as np
import scipy.integrate as si
import matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
# noinspection PyUnresolvedReferences
from mpl_toolkits.mplot3d.axes3d import Axes3D

from qspec.qtypes import *
from qspec import tools
import qspec.algebra as al
from qspec.physics import f_recoil, saturation
from qspec.simulate._simulate_cpp import sr_generate_y, Polarization, Environment, Atom, Laser

__all__ = ['ct_markov_analytic', 'ct_markov_dgl', 'lambda_states', 'lambda_ge_rec', 'Geometry', 'ScatteringRate']


try:
    matplotlib.use('Qt5Agg')
except ImportError:
    pass
# config = tools.get_config_dict()
# plt.rcParams['animation.ffmpeg_path'] = config['ffmpeg']

LINESTYLES = ['-', ':', '--', '-.']


def ct_markov_analytic(t: array_like, n: array_like, rates: array_like, r: array_like = 1.):
    """
    Computes the analytic solution of the evolution of a linear continuous-time markov chain.

    CAUTION! Computation of analytic solution problematic
    (Multiplication of very small numbers with very large numbers).

    :param t: The time at which the probability is returned (us).
    :param n: The number of the state for which the probability is returned. The first state corresponds to 'n' = 0.
    :param rates: The rate(s) at which the markov chain is transferring population from state i to state i + 1.
     If 'rates' is a scalar, this rate is assumed for all 0 &#8804; i &#8804; 'n'. If 'rates' is an Iterable,
     it must have length max('n') + 1 and individual rates 'rates[i]' are assumed for the states
     0 &#8804; i &#8804; 'n'.
    :param r: The ratio of the population which is transferred from state i to state i + 1. Then 1 - r is the ratio
     of the population that is lost during the transition from state i to state i + 1.
    :returns: The probability for the markov chain to be in state 'n' after the time 't' for P(t=0, n=0) = 1.
     The normalization condition is sum_n(P(n, t0)) = 1 for every single point in time t0.
     If 't' and 'n' are Iterables, a 2-d array is returned
     with shape (t.size, n.size), containing all probabilities P(t, n).
    """
    t = np.asarray(t)
    t_scalar = False
    if len(t.shape) == 0:
        t_scalar = True
    t = np.expand_dims(t.flatten(), axis=(1, 2))

    n = np.asarray(n, dtype=int)
    n_zero = n == 0
    n_max = np.max(n)
    n_scalar = False
    if len(n.shape) == 0:
        n_scalar = True
        # return np.array([ct_markov(t, n_i, rates[:(int(n_i)+1)], r=r) for n_i in n.flatten()]).T
    n = np.expand_dims(n.flatten(), axis=(0, 1))

    rates = np.asarray(rates)
    r = np.asarray(r)
    if len(r.shape) > 0 or np.abs(r - 0.5) > 0.5:
        raise ValueError('\'r\' must be a scalar between 0 and 1.')

    if len(rates.shape) == 0:
        ret = (r * rates * t[:, 0, :]) ** n[:, 0, :] / tools.factorial(n[:, 0, :]) * np.exp(-rates * t[:, 0, :])
    else:
        b = np.full((n_max, n_max), np.nan)

        def _b(_i, _n):
            _b_k = b[_i, _n - 1]
            if not np.isnan(_b_k):
                return _b_k
            if _i == _n - 1:
                if _n == 1:
                    return r * rates[0] / (rates[1] - rates[0])
                return r * rates[_i] / (rates[_i] - rates[_n]) * np.sum([_b(_j, _i) for _j in range(_n - 1)])
            elif _i <= _n - 2:
                return r * rates[_n - 1] / (rates[_n] - rates[_i]) * _b(_i, _n - 1)
            return 0.

        for n_i in range(1, n_max + 1, 1):
            for i_i in range(n_i):
                b[i_i, n_i - 1] = _b(i_i, n_i)

        b[np.isnan(b)] = 0.
        b = np.expand_dims(b, axis=0)
        b_map = n.flatten()[~n_zero] - 1

        ret = np.zeros((t.size, n.size))
        ret[:, n_zero] = np.exp(-rates[0] * t[:, 0, :])
        rates = np.expand_dims(rates, axis=(0, 2))

        ret[:, ~n_zero] = np.sum(b[:, :, b_map] * np.exp(-rates[:, :-1, :] * t), axis=1) \
            - np.sum(b[:, :, b_map], axis=1) \
            * np.exp(-np.transpose(rates[:, b_map + 1, :], axes=[0, 2, 1]) * t[:, 0, :])

    if t_scalar:
        if n_scalar:
            return ret[0, 0]
        return ret[0, :]
    if n_scalar:
        return ret[:, 0]
    return ret


def ct_markov_dgl(t: array_like, n: int, rates: array_like, r: array_like = 1., p0: array_like = None,
                  time_resolved: bool = False, show: bool = False):
    """
    Computes the evolution of a linear continuous-time markov chain numerically by solving the underlying ODE system.

    :param t: The time after which the probability is returned.
     If all times from 0 to 't' are required, use 'time_resolved'=True (us).
    :param n: The maximum state number to compute. The first state corresponds to 'n' = 0.
    :param rates: The rate(s) at which the markov chain is transferring population from state i to state i + 1.
     If 'rates' is a scalar, this rate is assumed for all 0 &#8804; i &#8804; 'n'. If 'rates' is an Iterable,
     individual rates 'rates[i]' are assumed for the states i and 'rates' must have size 'n' + 1.
    :param r: The ratio of the population which is transferred from state i to state i + 1. Here 1 - r is the ratio
     of the population that is lost during the transition from state i to state i + 1.
    :param p0: The initial population distribution. Must be an Iterable of length max('n') + 1.
     If None, 'p0' is set to [1, 0, 0, ..., 0].
    :param time_resolved: Whether to return the complete history of the result.
     If True, a 2-tuple similar to (time, population) is returned, were population has shape (time.size, n + 1).
     time will be an array of equally spaced times, such that numerical integrations can be performed easily.
    :param show: Whether to plot the result.
    :returns: The probability distribution of the states 0 &#8804; i &#8804; 'n' after the time 't'
     for P(t=0, n=i) = p0[i] The normalization condition is sum_n(P(n, t0)) = 1 for every single point in time t0.
    """
    t = np.asarray(t, dtype=float)
    if len(t.shape) > 0:
        raise ValueError('\'t\' must be a scalar.')
    n_dim = int(n) + 1
    rates = np.asarray(rates)
    if len(rates.shape) == 0:
        rates = np.full(n_dim, rates)
    else:
        rates = rates.flatten()
        if rates.size != n_dim:
            raise ValueError('\'rates\' must have size max(\'n\') + 1, but has size {}.'.format(rates.size))
    r = np.asarray(r)
    if len(r.shape) > 0 or np.abs(r - 0.5) > 0.5:
        raise ValueError('\'r\' must be a scalar between 0 and 1.')

    if p0 is None:
        p0 = np.zeros(n + 1)
        p0[0] = 1.
    else:
        p0 = np.asarray(p0, dtype=float).flatten()
        if p0.size != n_dim:
            raise ValueError('\'p0\' must have size max(\'n\') + 1, but has size {}.'.format(p0.size))

    def _f(_t, _y):
        return np.array([-rates[0] * _y[0], ] + [r * _ri * _yi - _rj * _yj for _ri, _yi, _rj, _yj
                                                 in zip(rates[:-1], _y[:-1], rates[1:], _y[1:])])

    jac = np.zeros((n_dim, n_dim))
    for i, ri in enumerate(rates):
        jac[i, i] = -ri
        if i > 0:
            jac[i, i - 1] = r * rates[i - 1]

    def _df(_t, _y):
        return jac

    dt = 0.05 / np.max(rates)
    t_array = np.linspace(0., t, tools.odd(t / dt))
    ode = si.ode(_f, jac=_df)
    ode.set_initial_value(p0, 0.)
    # noinspection PyTypeChecker
    y_array = np.array([p0, ] + [ode.integrate(ti) for ti in t_array[1:]])
    # result = si.solve_ivp(_f, (0., t), p0, method='RK45', max_step=dt, t_eval=t_array, vectorized=True)
    # # noinspection PyUnresolvedReferences
    # t_array, y_array = result.t, result.y.T

    if show:
        for ni, yi in enumerate(y_array.T):
            plt.plot(t_array, yi, label='n={}'.format(ni))
        plt.plot(t_array, np.sum(y_array, axis=1), 'k--', label=r'$N$')
        if y_array.shape[1] < 10:
            plt.legend()
        plt.show()

    if time_resolved:
        return t_array, y_array
    return y_array[-1, :]


def lambda_states(t: array_like, delta_1: array_like, delta_2: array_like, a_ge: array_like, a_me: array_like,
                  s_1: array_like, s_2: array_like, lw_1: array_like = 0., lw_2: array_like = 0., p0: array_like = None,
                  time_resolved: bool = False, show: bool = False):
    """
    Computes the evolution of Lambda-systems such as the alkali metals or the singly-charged alkaline-earth metals.
    The state vector is defined as (g, m, e), where g is the first end of the Lambda,
    m the second end and e the intermediate state.

    :param t: The time after which the probability is returned.
     If all times from 0 to 't' are required, use 'time_resolved'=True (us).
    :param delta_1: The detuning of the first laser relative to the g->e transition.
    :param delta_2: The detuning of the second laser relative to the m->e transition.
    :param a_ge: The Einstein coefficient of the e->g transition.
    :param a_me: The Einstein coefficient of the e->m transition.
    :param s_1: The saturation parameter of the g->e transition.
    :param s_2: The saturation parameter of the m->e transition.
    :param lw_1: The frequency width of the first laser.
    :param lw_2: The frequency width of the second laser.
    :param p0: The initial density matrix. Must have shape (6, ), containing the elements [gg, mm, ee, gm, eg, em].
     If None, initially all population is in the g state.
    :param time_resolved: Whether to return the complete history of the result.
    :param show: Whether to plot the result.
    :returns: The density matrix elements after the time 't'. If time_resolved is True,
     a 2-tuple similar to (time, density matrix) is returned, were the density matrix has shape (time.size, 6).
     time will be an array of equally spaced times, such that numerical integrations can be performed easily.
    """
    t = np.asarray(t, dtype=float)
    if len(t.shape) > 0:
        raise ValueError('\'t\' must be a scalar.')

    if p0 is None:
        p0 = np.zeros(6, dtype=complex)
        p0[0] = 1.
    else:
        p0 = np.asarray(p0, dtype=complex).flatten()
        if p0.size != 6:
            raise ValueError('\'p0\' must have size 6, but has size {}.'.format(p0.size))

    _delta_1 = 2 * np.pi * delta_1
    _delta_2 = 2 * np.pi * delta_2
    rabi_1 = a_ge * np.sqrt(s_1 / 2.)
    rabi_2 = a_me * np.sqrt(s_2 / 2.)

    def _f(_t, _y):
        i_gg = _y[0]
        i_mm = _y[1]
        i_ee = _y[2]

        i_gm = _y[3]
        i_eg = _y[4]
        i_em = _y[5]

        gg = a_ge * i_ee + 1j * rabi_1 * (np.conj(i_eg) - i_eg) / 2.
        mm = a_me * i_ee + 1j * rabi_2 * (np.conj(i_em) - i_em) / 2.
        ee = -(a_ge + a_me) * i_ee - 1j * rabi_1 * (np.conj(i_eg) - i_eg) / 2. \
            - 1j * rabi_2 * (np.conj(i_em) - i_em) / 2.

        gm = (1j * (_delta_2 - _delta_1) - (lw_1 + lw_2) / 2.) * i_gm \
            + 1j * rabi_2 * np.conj(i_eg) / 2. - 1j * rabi_1 * i_em / 2.
        eg = (1j * _delta_1 - (a_ge + a_me + lw_1) / 2.) * i_eg \
            + 1j * rabi_1 * (i_ee - i_gg) / 2. - 1j * rabi_2 * np.conj(i_gm) / 2.
        em = (1j * _delta_2 - (a_ge + a_me + lw_2) / 2.) * i_em \
            + 1j * rabi_2 * (i_ee - i_mm) / 2. - 1j * rabi_1 * i_gm / 2.

        return np.array([gg, mm, ee, gm, eg, em], dtype=complex)

    dt = 0.1 / np.max([a_ge + rabi_1 + lw_1, a_me + rabi_2 + lw_2])
    t_array = np.linspace(0., t, tools.odd(t / dt))
    ode = si.complex_ode(_f)
    ode.set_initial_value(p0, 0.).set_integrator('vode', method='bdf')
    # noinspection PyTypeChecker
    y_array = np.array([p0, ] + [ode.integrate(ti) for ti in t_array[1:]])
    # result = si.solve_ivp(_f, (0., t), p0, method='RK45', max_step=dt, t_eval=t_array, vectorized=True)
    # # noinspection PyUnresolvedReferences
    # t_array, y_array = result.t, result.y.T
    if show:
        plt.plot(t_array, y_array[:, 0].real, label=r'$\rho_\mathrm{gg}$')
        plt.plot(t_array, y_array[:, 1].real, label=r'$\rho_\mathrm{mm}$')
        plt.plot(t_array, y_array[:, 2].real, label=r'$\rho_\mathrm{ee}$')
        plt.legend(loc=5)
        plt.show()

    if time_resolved:
        return t_array, y_array
    return y_array[-1, :]


def lambda_ge_rec(t: array_like, n: array_like, delta: array_like, a_ge: array_like, a_me: array_like, s: array_like,
                  f: array_like, m: array_like, p0: array_like = None, dt: float = None, time_resolved: bool = False,
                  show: bool = False):
    """
    Computes the evolution of Lambda-systems such as the alkali metals or the singly-charged alkaline-earth metals,
    taking into account photon recoils. The system is driven by a single laser.
    The state vector is defined as (g0, m0, e0, g1, m1, e1, ..., gn, mn, en), where g is the first end of the Lambda,
    m the second end and e the intermediate state.
    The photon recoils are modeled using a discrete subspace and not the continuous momentum space.
    The number of photon recoils increases by one when the system is excited from the g state to the e state.
    Vice-versa, the number of photon recoils decreases by one when the system is deexcited
    from the e state to the g state via the process of stimulated emission. When the system decays into the g state
    via the dissipative mechanism described by the Einstein coefficient 'a_ge',
    the number of photon recoils does not change.

    :param t: The time after which the probability is returned.
     If all times from 0 to 't' are required, use 'time_resolved'=True (us).
    :param n: The maximum number of photon recoils to consider.
    :param delta: The detuning of the laser relative to the g->e transition (MHz).
    :param a_ge: The Einstein coefficient of the e->g transition (MHz).
    :param a_me: The Einstein coefficient of the e->m transition (MHz).
    :param s: The saturation parameter of the g->e transition.
    :param f: The transition frequency of the g->e transition (MHz).
    :param m: The mass number of the element (u).
    :param p0: The initial density matrix. Must have shape (6, ), containing the elements [gg, mm, ee, gm, eg, em](0)
     or be the full density matrix with all the recoil information.
    :param dt: The width of the time steps.
    :param time_resolved: Whether to return the complete history of the result.
    :param show: Whether to plot the result.
    :returns: The diagonal density matrix elements after the time 't'. If time_resolved is True,
     a 2-tuple similar to (time, density matrix) is returned, were the density matrix has shape (time.size, 3(n+1)).
     time will be an array of equally spaced times, such that numerical integrations can be performed easily.
    """
    t = np.asarray(t, dtype=float)
    if len(t.shape) > 0:
        raise ValueError('\'t\' must be a scalar.')
    if p0 is None:
        y0 = np.zeros(int(3 * (n + 1) * (3 * (n + 1) + 1) / 2), dtype=complex)
        y0[0] = 1. + 0.j
    else:
        p0 = np.asarray(p0, dtype=float).flatten()
        if p0.size == 6:
            y0 = np.zeros(int(3 * (n + 1) * (3 * (n + 1) + 1) / 2), dtype=complex)
            y0[:6] = p0
        elif p0.size == int(3 * (n + 1) * (3 * (n + 1) + 1) / 2):
            y0 = p0
        else:
            raise ValueError('\'p0\' must have size {}, but has size {}.'
                             .format(int(3 * (n + 1) * (3 * (n + 1) + 1) / 2), p0.size))

    _delta = 2 * np.pi * delta
    rabi = a_ge * np.sqrt(s / 2.)
    f_rec = 2 * np.pi * f_recoil(f, m)

    def hamiltonian(_t, n1, n2) -> ndarray:  # without hbar
        gg, mm, ee = 0. + 0.j, 0. + 0.j, 0. + 0.j
        gm, eg, me = 0. + 0.j, 0. + 0.j, 0. + 0.j
        mg, ge, em = 0. + 0.j, 0. + 0.j, 0. + 0.j
        if n1 == n2:
            _e_kin = n1 ** 2 * f_rec
            # _delta includes one recoil, _delta := (w_eg - w_L +- k*v + k*v_recoil)
            gg = -(_delta + 2 * n1 * f_rec) + _e_kin
            mm, ee = _e_kin, _e_kin
        elif n1 == n2 - 1:
            ge = rabi / 2.
        elif n1 == n2 + 1:
            eg = rabi / 2.
        else:
            return np.zeros((3, 3), dtype=complex)
        return np.array([[gg, gm, ge], [mg, mm, me], [eg, em, ee]], dtype=complex)

    def rho(_y):
        ret = np.zeros((3 * (n + 1), 3 * (n + 1)), dtype=complex)
        np.fill_diagonal(ret, _y[:(3 * (n + 1))])
        for _i in range(3 * (n + 1) - 1):
            i_start = 3 * (n + 1) + sum([3 * (n + 1) - _j - 1 for _j in range(_i)])
            i_end = i_start + sum([3 * (n + 1) - _j - 1 for _j in range(_i + 1)])
            np.fill_diagonal(ret[(_i+1):], _y[i_start:i_end].conjugate())
            np.fill_diagonal(ret[:, (_i+1):], _y[i_start:i_end])
        return ret

    def _f(_t, _y):
        _rho = rho(_y)
        h = np.block(np.array([[hamiltonian(_t, n1, n2) for n2 in range(n + 1)] for n1 in range(n + 1)]))
        ret = -1.j * (h @ _rho - _rho @ h)

        zero = np.zeros((3, 3))
        sigma = np.array([[0., 0., 0.], [0., 0., 0.], [1., 0., 0.]])
        sigma_eg = np.block(np.array([[sigma if _i == _j else zero for _j in range(n + 1)] for _i in range(n + 1)]))
        sigma = np.array([[0., 0., 0.], [0., 0., 0.], [0., 1., 0.]])
        sigma_em = np.block(np.array([[sigma if _i == _j else zero for _j in range(n + 1)] for _i in range(n + 1)]))
        lindbladian_ge = sigma_eg.T @ _rho @ sigma_eg \
            - 0.5 * (sigma_eg @ sigma_eg.T @ _rho + _rho @ sigma_eg @ sigma_eg.T)
        lindbladian_me = sigma_em.T @ _rho @ sigma_em \
            - 0.5 * (sigma_em @ sigma_em.T @ _rho + _rho @ sigma_em @ sigma_em.T)
        lindbladian_ge *= a_ge
        lindbladian_me *= a_me
        ret += lindbladian_ge + lindbladian_me
        return np.concatenate(tuple(np.diagonal(ret, offset=_i) for _i in range(3 * (n + 1))), axis=0)

    dt = 0.1 / (a_ge + rabi) if dt is None else dt
    t_array = np.linspace(0., t, tools.odd(t / dt))
    ode = si.complex_ode(_f)
    ode.set_initial_value(y0, 0.).set_integrator('vode', method='bdf')
    # noinspection PyTypeChecker
    y_array = np.array([y0, ] + [ode.integrate(ti) for ti in t_array[1:]])
    # result = si.solve_ivp(_f, (0., t), y0, method='RK45', max_step=dt, t_eval=t_array, vectorized=True)
    # # noinspection PyUnresolvedReferences
    # t_array, y_array = result.t, result.y.T
    if show:
        plt.plot(t_array, np.sum([y_array[:, 3 * i].real for i in range(n + 1)], axis=0),
                 'C0--', label=r'$N_\mathrm{gg}$')
        plt.plot(t_array, np.sum([y_array[:, 3 * i + 1].real for i in range(n + 1)], axis=0),
                 'C1--', label=r'$N_\mathrm{mm}$')
        plt.plot(t_array, np.sum([y_array[:, 3 * i + 2].real for i in range(n + 1)], axis=0),
                 'C2--', label=r'$N_\mathrm{ee}$')
        plt.plot(t_array, np.sum([y_array[:, 3 * i].real + y_array[:, 3 * i + 1].real + y_array[:, 3 * i + 2].real
                                  for i in range(n + 1)], axis=0), 'k--', label=r'$N$')
        pg, pm, pe = None, None, None
        for ni, yi in enumerate(y_array.T[:(3 * (n + 1))]):
            if ni % 3 == 0:
                pg, = plt.plot(t_array, yi.real, 'b-')
            if ni % 3 == 1:
                pm, = plt.plot(t_array, yi.real, 'r-')
            if ni % 3 == 2:
                pe, = plt.plot(t_array, yi.real, 'g-')
        pg.set_label(r'$\rho_\mathrm{gg}^n$')
        pm.set_label(r'$\rho_\mathrm{mm}^n$')
        pe.set_label(r'$\rho_\mathrm{ee}^n$')
        plt.xlabel('Time (us)')
        plt.xlabel('Population')
        plt.legend(loc=7)
        plt.show()

        plt.bar(np.arange(n + 1), [y_array[-1, 3 * i + 0].real for i in range(n + 1)])
        plt.show()

    if time_resolved:
        return t_array, y_array[:, :(3 * (n + 1))].real
    return y_array[-1, :(3 * (n + 1))].real


class Geometry:
    def __init__(self):
        r"""
        Class representing a fluorescence detection geometry. The solid angle over which fluorescence light is detected
        can be defined through intervals of the two angles $\theta$ and $\phi$.
        With these, every spacial direction can be addressed using an orthonormal system defined by

        $$
        \hat{e}_r = \begin{pmatrix}\sin(\theta)\\
                    \cos(\theta)\sin(\phi)\\
                    \cos(\theta)\cos(\phi)\end{pmatrix}.
        $$

        If the user specifies a rotation object with unitary matrix $R$,
        the new system is $\hat{e}_r^\prime = R \hat{e}_r$
        The entire two-dimensional interval is defined through the cartesian product
        $\bigcup_i \theta_i \times \bigcup_i \phi_i$.
        For every disjoint interval a weight can be defined through a 'weights' matrix.
        A probability distribution function (pdf) can be defined to have continuous angle weights.
        A rotation matrix can be defined to rotate the entire coordinate systems/detection geometry.
        A sample of angle pairs from the defined intervals can be generated using the 'integration_sample' method.
        """
        self.theta_intervals = np.array([[-np.pi / 2., np.pi / 2.]])
        self.phi_intervals = np.array([[0., 2. * np.pi]])
        self.solid_angle = 4 * np.pi
        self.step = np.pi / 32.
        self.pdf = None
        self.weights = None
        self.rotation = tools.Rotation()

    def set_intervals(self, theta: array_iter, phi: array_iter):
        """
        :param theta: An interval or a list of intervals for the angle 'theta'.
        :param phi: An interval or a list of intervals for the angle 'phi'.
        :returns: None. Merges overlapping intervals and calculates the solid angle
         corresponding to the defined intervals.
        """
        theta, phi = np.asarray(theta), np.asarray(phi)
        if len(theta.shape) == 1:
            theta = np.expand_dims(theta, axis=0)
        if len(phi.shape) == 1:
            phi = np.expand_dims(phi, axis=0)
        if len(theta.shape) != 2 or len(phi.shape) != 2:
            raise ValueError('theta and phi must be intervals or lists of intervals.')
        self.theta_intervals = tools.merge_intervals(theta)
        self.phi_intervals = tools.merge_intervals(phi)
        y_int = 0.
        theta_int, phi_int = self.integration_sample()
        for t in theta_int:
            y = np.cos(t)
            y_int += si.simpson(y, x=t)
        y_int *= np.sum([p[-1] - p[0] for p in phi_int])
        self.solid_angle = y_int

    def set_weights(self, weights: Union[array_iter, None]):
        """
        :param weights: None or a matrix of weights for the defined disjoint intervals.
         The shape of the weights must fulfill weights.shape == (len(self.theta_intervals), len(self.phi_intervals)).
        :returns: None. Sets the 'weights' attribute of the geometry object.
        """
        if weights is not None:
            weights = np.asarray(weights)
            shape = (self.theta_intervals.shape[0], self.phi_intervals.shape[0])
            if weights.shape != shape:
                raise ValueError('Weights must have shape {}, but have shape {}.'.format(shape, weights.shape))
        self.weights = weights

    def set_pdf(self, pdf: Union[Callable, None]):
        """
        :param pdf: None or a callable which accepts two arguments (theta, phi).
        :returns: None. Sets the 'pdf' attribute of the Geometry object.
        """
        if pdf is not None:
            try:
                pdf(self.theta_intervals[0, 0], self.phi_intervals[0, 0])
            except AttributeError:
                raise AttributeError('The pdf must be a callable which takes two arguments *(theta, phi).')
        self.pdf = pdf

    def set_rotation(self, r: tools.Rotation = None):
        """
        :param r: None or a 3x3 matrix defining a rotation.Therefore, dot(r, r.T) = I and Det(r) = 1 must be fulfilled.
         If r == None, the Identity matrix is used.
        :returns: None. Sets the 'R' attribute of the Geometry object.
        """
        if r is None:
            r = tools.Rotation()
        elif not isinstance(r, tools.Rotation):
            raise TypeError('r must be a \'Rotation\' object but is of type {}'.format(type(r)))
        self.rotation = r

    def integration_sample(self, step: scalar = None) -> (list, list):
        """
        :param step: None or a scalar which defines the approximate spacing
         between the equidistant values of the integration sample.
         If step == None, the currently defined step size is used. The standard value is pi/32.
        :returns: Two lists of arrays with equidistant values
         in the defined intervals for 'theta' and 'phi', respectively.
         Note that all intervals are exactly covered by an odd number of values (for Simpson integration)
         under the expense of not matching the 'step' size exactly.
        """
        if step is None:
            step = self.step
        theta = [np.linspace(t[0], t[1], tools.odd((t[1] - t[0]) / step + 1.)) for t in self.theta_intervals]
        phi = [np.linspace(p[0], p[1], tools.odd((p[1] - p[0]) / step + 1.)) for p in self.phi_intervals]
        return theta, phi

    def plot(self, show: bool = True):
        """
        Shows a 3d plot of the angular range covered by the geometry object.
        :returns: The axes object.
        """

        fig = plt.figure(num=1, figsize=(8, 8), clear=True)
        ax = fig.add_subplot(111, projection='3d')
        plt.xlabel(r'$x$ / arb. units')
        plt.ylabel(r'$y$ / arb. units')
        ax.set_zlabel(r'$z$ / arb. units')

        lim = 1.1
        ax.set_xlim(-lim, lim)
        ax.set_ylim(-lim, lim)
        ax.set_zlim(-lim, lim)
        ticks = [-1., -0.5, 0., 0.5, 1.]
        ax.set_xticks(ticks)
        ax.set_yticks(ticks)
        ax.set_zticks(ticks)
        ax.pbaspect = [1., 1., 1.]

        radius = 0.9
        theta, phi = self.integration_sample(step=0.02)
        for t_arr in theta:
            for p_arr in phi:
                r = np.array([[tools.e_r(t, p) for p in p_arr] for t in t_arr])
                r = tools.transform(np.expand_dims(self.rotation.R, axis=(0, 1)), r)
                r *= radius
                surf = ax.plot_surface(r[:, :, 0], r[:, :, 1], r[:, :, 2], zorder=2, rcount=40, ccount=40,
                                       cmap=plt.get_cmap('cividis'), antialiased=False, alpha=0.5)
                surf.set_clim([-radius, radius])
        # cb = plt.colorbar(surf, ax=ax, shrink=0.5)
        # cb.set_label('z / arb. units')
        # cb.set_ticks([-1, 0, 1])

        # plt.style.use('seaborn')
        # plt.tight_layout()
        if show:
            plt.show()
            plt.style.use('default')

        return fig, ax


class ScatteringRate:
    """
    Class representing the perturbative scattering rate of a closed electronic transition.
    The spectrum is excited by a laser with the specified 'polarization'
    and recorded with a FDR defined by 'geometry'. An external magnetic field can be defined through 'b'.
    """
    def __init__(self, atom: Atom, laser: Laser = None, i_decay: int = 0, geometry: Geometry = None,
                 b: array_like = None):
        """
        :param atom: The investigated atom.
        :param laser: The incident laser beam.
        :param i_decay: The indexing of the transition in atom.decay_map.
        :param geometry: The geometry of the FDR.
        :param b: The external magnetic field. Can be a vector or a scalar.
         In the latter case, the magnetic field is aligned with the z-axis.
        """
        self.environment = Environment()
        self.atom = atom
        self.freq_0 = 0.
        self.laser, self.geometry, self.polarization = None, None, None
        self.i_decay = i_decay
        self.state_l, self.state_u = None, None
        self.i_state_l, self.i_state_u = None, None
        self.v = None
        self.s = 1.
        self.gamma = None
        self.A, self.a_const = None, None
        self.x0 = None
        self.indices = None
        self.x0_array, self.A_i, self.A_f, self.counts = None, None, None, None
        self.b, self.b_abs = np.array([0., 0., 0.]), 0.
        self.e_r_b = np.array([0., 0., 1.])
        self.e_l = np.array([0., 0., 1.])
        self.R_b, self.R = np.identity(3, dtype=float), np.identity(3, dtype=float)
        self.set_states()
        self.check_atom()
        self.set_laser(laser)
        self.set_geometry(geometry)
        self.generate_dipoles()
        self.set_b(b)  # TODO: Implement Environments in C++.
    
    def check_atom(self):
        """
        Check whether the structure of the atom fulfills all requirements to calculate the scattering rate.

        :raises ValueError: The atom has to consist of a single closed transition between two fine-structure states.
         Their common nucleus can have an arbitrary spin.
        :returns: None.
        """
        s, l, j, i = self.state_l[0].s, self.state_l[0].l, self.state_l[0].j, self.state_l[0].i
        if any(state_l.s != s or state_l.l != l or state_l.j != j or state_l.i != i for state_l in self.state_l):
            raise ValueError('All lower states must be part of the same fine-structure state.')
        s, l, j, i = self.state_u[0].s, self.state_u[0].l, self.state_u[0].j, self.state_u[0].i
        if any(state_u.s != s or state_u.l != l or state_u.j != j or state_u.i != i for state_u in self.state_u):
            raise ValueError('All upper states must be part of the same fine-structure state.')
        if self.state_l[0].i != self.state_u[0].i:
            raise ValueError('The lower and upper states must be part of the same nuclear state.')

    def set_geometry(self, geometry: Geometry = None):
        """
        :param geometry: The geometry object. If None, the entire solid angle is covered by the detector.
        :return: None. Sets the geometry object.
        """
        self.geometry = geometry
        if self.geometry is None:
            self.geometry = Geometry()
        self.R = np.dot(self.geometry.rotation.R, self.R_b)

    def set_polarization(self, polarization: Polarization = None):
        """
        :param polarization: The polarization object. If None, the light is linearly polarized along the z-axis.
        :returns: None. Sets the polarization object and the local polarization vector 'e_l'.
        """
        self.polarization = polarization
        if self.polarization is None:
            self.polarization = Polarization([0, 0, 1])
        self.e_l = np.dot(self.R_b, self.polarization.x)
        self.polarization.def_q_axis(self.e_r_b)

    def set_laser(self, laser: Laser = None):
        self.laser = laser
        if self.laser is None:
            self.laser = Laser(self.freq_0, intensity=1., polarization=Polarization())
        self.s = saturation(self.laser.intensity, self.laser.freq, self.atom.decay_map.a[self.i_decay], 1.)
        self.set_polarization(self.laser.polarization)

    def set_states(self):
        """
        Define lower and upper states.

        :returns: None.
        """
        state_l = np.array([state for state in self.atom if state.label == self.atom.decay_map.labels[self.i_decay][0]])
        state_u = np.array([state for state in self.atom if state.label == self.atom.decay_map.labels[self.i_decay][1]])
        self.freq_0 = np.mean([s_u.freq_j for s_u in state_u]) - np.mean([s_l.freq_j for s_l in state_l])
        if self.freq_0 > 0:
            self.state_u = state_u
            self.state_l = state_l
        else:
            self.freq_0 *= -1
            self.state_u = state_l
            self.state_l = state_u

    def set_x0(self):
        """
        :returns: None. Updates the resonance positions.
        """
        self.x0 = [[state_u.freq - state_l.freq - self.laser.freq
                    for state_u in self.state_u] for state_l in self.state_l]
        self.x0_array = np.array([self.x0[col[0]][col[1]] for row in self.indices for col in row], dtype=float)

    def generate_dipoles(self):
        """
        :returns: None. Generates an array of frequencies which cover the entire spectrum
         and saves it to the x attribute of the Spectrum object.
        """
        i = self.state_l[0].i
        self.gamma = self.atom.decay_map.a[self.i_decay] / (2 * np.pi)
        j_l = self.state_l[0].j
        j_u = self.state_u[0].j

        self.A = [[np.zeros(3, dtype=complex) if abs(u.f - ll.f) > 1 or abs(u.m - ll.m) > 1
                   else al.a_dipole_cart(i, j_l, ll.f, ll.m, j_u, u.f, u.m)
                   for u in self.state_u] for ll in self.state_l]

        self.a_const = [[0. if abs(u.f - ll.f) > 1 or abs(u.m - ll.m) > 1
                         else al.a_dipole(i, j_l, ll.f, ll.m, j_u, u.f, u.m, u.m - ll.m, as_sympy=False)
                         for u in self.state_u] for ll in self.state_l]

        self.indices = [[[i_i, i_u, i_f] for i_u, u in enumerate(self.state_u)
                         if abs(u.f - i.f) <= 1 and abs(u.m - i.m) <= 1
                         and abs(u.f - f.f) <= 1 and abs(u.m - f.m) <= 1]
                        for i_i, i in enumerate(self.state_l) for i_f, f in enumerate(self.state_l)]
        self.indices = [arg for arg in self.indices if arg]
        self.A_i = np.array([self.A[col[0]][col[1]] for row in self.indices for col in row], dtype=complex)
        self.A_f = np.array([self.A[col[2]][col[1]] for row in self.indices for col in row], dtype=complex)
        self.counts = np.array([len(row) for row in self.indices], dtype=int)
        self.set_x0()

    def set_b(self, b: array_like):
        """
        :param b: The magnetic field vector.
        :returns: None. Sets the new magnetic field for all states
         and updates the quantization axis and the resonance positions.
        """
        if b is None:
            self.b = np.array([0., 0., 0.])
            self.b_abs = 0.
            self.e_r_b = np.array([0., 0., 1.])
        else:
            self.b = np.asarray(b)
            if len(self.b.shape) == 0:
                self.b = np.array([0., 0., self.b])
            self.b_abs = tools.absolute(self.b)
            self.e_r_b = np.array([0., 0., 1.])
            if not self.b_abs == 0.:
                self.e_r_b = self.b / self.b_abs
        z_axis = np.array([0., 0., 1.])
        self.R_b = tools.rotation_to_vector(self.e_r_b, z_axis).R
        self.R = np.dot(self.geometry.rotation.R, self.R_b)
        self.e_l = np.dot(self.R_b, self.polarization.x)
        self.polarization.def_q_axis(self.e_r_b)
        self.environment = Environment(B=self.b)
        for state in self.atom:
            state.update(self.environment)
        self.set_x0()

    def generate_x(self, width: scalar = 20., step: scalar = None):
        """
        :param width: The covered width around resonances in natural linewidths.
        :param step: The step size between generated x values.
         Note that between resonances, there might be a larger step.
        :returns: An array of frequencies which covers the entire spectrum.
        """
        intervals = [[x0 - width * self.gamma / 2, x0 + width * self.gamma / 2] for x0 in self.x0_array]
        intervals = tools.merge_intervals(intervals)
        if step is None:
            step = self.gamma / 30.
        return np.concatenate([np.linspace(i[0], i[1], tools.odd((i[1] - i[0]) / step + 1.))
                               for i in intervals], axis=0)

    def generate_y0(self, x: array_like) -> array_like:
        """
        :param x: The frequency of light in an atoms rest frame (MHz).
        :returns: The spectrum for the case of unpolarized light
         and the detection of the complete solid angle (4 pi).
        """
        y0 = np.zeros_like(x, dtype=float)
        for x0_list, a_const_list in zip(self.x0, self.a_const):
            for x0, a_const in zip(x0_list, a_const_list):
                denominator = (x - x0) ** 2 + 0.25 * self.gamma ** 2
                y0 += a_const ** 2 / denominator
        return y0 * (self.gamma / 2.) ** 3 / len(self.x0) / 3 * 2 * np.pi * self.s

    def generate_y_4pi(self, x: array_like) -> array_like:
        """
        :param x: The frequency of light in an atoms rest frame (MHz).
        :returns: The spectrum for the detection of the complete solid angle (4 pi).
        """
        y0 = np.zeros_like(x, dtype=float)
        for x0_list, a_const_list, ll in zip(self.x0, self.a_const, self.state_l):
            for x0, a_const, u in zip(x0_list, a_const_list, self.state_u):
                if abs(u.f - ll.f) > 1 or abs(u.m - ll.m) > 1:
                    continue
                denominator = (x - x0) ** 2 + 0.25 * self.gamma ** 2
                y0 += a_const ** 2 / denominator * tools.absolute_complex(self.polarization.q[int(u.m - ll.m) + 1]) ** 2
        return y0 * (self.gamma / 2.) ** 3 / len(self.x0) * 2 * np.pi * self.s

    def generate_y(self, x: array_like, theta: array_like, phi: array_like, decimals: int = 8):
        """
        :param x: The frequency of light in an atoms rest frame (MHz).
        :param theta: The angle between the emission direction of the fluorescence light and the x-axis + 90 deg.
        :param phi: The mixing angle between the y- and z-axis (sin(phi), cos(phi)).
        :param decimals: The precision of the vector calculus in considered decimal places.
        :returns: The fluorescence spectrum for incident light with frequency 'x'
         for the direction of emission defined by 'theta' and 'phi'.
        """
        x, theta, phi = np.asarray(x).flatten(), np.asarray(theta).flatten(), np.asarray(phi).flatten()
        t, p = np.meshgrid(theta, phi, indexing='ij')
        t, p = t.flatten(), p.flatten()
        e_theta = np.around(tools.transform(np.expand_dims(self.R, axis=0), tools.e_theta(t, p)), decimals=decimals)
        e_phi = np.around(tools.transform(np.expand_dims(self.R, axis=0), tools.e_phi(t, p)), decimals=decimals)
        e_theta /= np.expand_dims(tools.absolute(e_theta), axis=-1)
        e_phi /= np.expand_dims(tools.absolute(e_phi), axis=-1)
        # Lorentz Boost TODO
        # if self.v is not None:
        #     e_theta = boost(np.concatenate([[1.], e_theta], axis=0), self.v)[1:]
        #     e_phi = boost(np.concatenate([[1.], e_phi], axis=0), self.v)[1:]
        #     e_theta /= tools.absolute(e_theta)
        #     e_phi /= tools.absolute(e_phi)

        norm = 3. / (8. * np.pi) * (self.gamma / 2.) ** 3 / len(self.x0) * 2 * np.pi * self.s
        denominator = 1. / (-np.expand_dims(x, axis=1) + self.x0_array + 0.5j * self.gamma)

        i_l = np.around(np.sum(self.e_l * self.A_i, axis=-1), decimals=decimals)
        f_theta = np.around(np.sum(np.expand_dims(e_theta, axis=1)
                                   * np.expand_dims(self.A_f, axis=0), axis=-1) * i_l,
                            decimals=decimals)
        f_phi = np.around(np.sum(np.expand_dims(e_phi, axis=1)
                                 * np.expand_dims(self.A_f, axis=0), axis=-1) * i_l,
                          decimals=decimals)
        shape = np.array([denominator.shape[0], f_theta.shape[0], self.counts.shape[0]], dtype=int)
        denominator, f_theta, f_phi = denominator.flatten(), f_theta.flatten(), f_phi.flatten()
        t0 = time()
        ret = sr_generate_y(denominator, f_theta, f_phi, self.counts, shape).reshape((x.size, theta.size, phi.size))
        t0 = time() - t0
        # print('Time: {}s'.format(t0))
        return ret * norm

    def integrate_y(self, x: array_like, step: scalar = None):
        """
        :param x: The frequency of light in an atoms rest frame (MHz).
        :param step: The step size used for the integration over the angles theta and phi.
        :returns: The fluorescence spectrum for incident light with frequencies 'x'
         integrated over the directions of emission 'theta' and 'phi'.
        """
        x = np.asarray(x)
        theta, phi = self.geometry.integration_sample(step=step)
        y_int = np.zeros(x.shape)
        for i, t in enumerate(theta):
            for j, p in enumerate(phi):
                y = self.generate_y(x, t, p) * np.expand_dims(np.expand_dims(np.cos(t), axis=0), axis=-1)
                if self.geometry.pdf is not None:
                    y *= np.expand_dims(self.geometry.pdf(t, p), axis=0)
                y = si.simpson(y, x=t, axis=1)
                y = si.simpson(y, x=p)
                if self.geometry.weights is not None:
                    y *= self.geometry.weights[i, j]
                y_int += y
        return y_int

    def plot_spectrum(self, norm_to_4pi: bool = False, step: scalar = None):
        """

        :param norm_to_4pi: Whether to renormalize the spectrum defined by geometry to have the same maximum
         as the spectrum for the detection with the complete solid angle 4 pi.
         If True, the difference between the two spectra is plotted as well. Default is False.
        :param step: The step size for the integration of the scattering rate.
        :returns: None. Plots the spectrum for the given geometry
         and that for the detection with the complete solid angle 4 pi.
        """
        plt.figure(figsize=(8, 4))
        x = self.generate_x()
        y_4pi = self.generate_y_4pi(x)
        y = self.integrate_y(x, step=step)
        if norm_to_4pi:
            y *= np.max(y_4pi) / np.max(y)
        plt.xlabel(r'$\nu - \nu_0$ (MHz)')
        plt.ylabel(r'Scattering rate (MHz / sat. parameter)')
        plt.plot(x, y_4pi, label=r'$y(4\pi)$')
        plt.plot(x, y, label=r'$y($Geometry$)$')
        if norm_to_4pi:
            plt.plot(x, y - y_4pi, label='Residuals')
        plt.legend()
        plt.tight_layout()
        plt.show()

    def plot_angular_distribution(self, x: scalar = None, n: int = 5, theta: array_iter = None, phi: array_iter = None,
                                  mode: str = None, show: bool = True, save: str = None):
        """
        :param x: The frequency of light in an atoms rest frame (MHz).
         If None, the position of the first maximum is taken (self.x0[0][0]).
        :param n: 2 ** n + 1 samples for 'theta' and 2 ** (n + 1) + 1 samples for 'phi' are drawn,
         giving a total of 2 ** (2n + 1) + 3 * 2 ** n + 1 points to draw.
        :param theta: A custom array for theta. Overwrites 'n'.
        :param phi: A custom array for phi. Overwrites 'n'.
        :param mode: Either '3d' for a surface plot or anything else for a color-mesh plot.
        :param save: The path where to save the plot. If None, the plot will not be saved.
        :param show: Whether to show the plot.
        :returns: theta, phi and z. Optionally, a plot is drawn and saved.
        """
        if theta is not None:
            theta = np.asarray(theta)
        else:
            theta = np.linspace(-np.pi / 2., np.pi / 2., 2 ** n + 1)
        if phi is not None:
            phi = np.asarray(phi)
        else:
            phi = np.linspace(0., 2. * np.pi, 2 ** (n + 1) + 1)
        y = np.squeeze(self.generate_y(self.x0[0][0] if x is None else x, theta, phi))

        if mode is None:
            mode = '2d'
        if mode == '3d':
            r = np.array([[tools.transform(self.geometry.rotation.R, tools.e_r(t, p)) for p in phi] for t in theta])
            fig = plt.figure(num=1, figsize=(8, 8), clear=True)
            ax = fig.add_subplot(111, projection='3d', label='Surface')
            rad = y[:, :]
            rad = np.expand_dims(rad, axis=-1)
            r *= rad
            r -= np.min(tools.absolute(r, axis=-1)) * (r / np.expand_dims(tools.absolute(r, axis=-1), axis=-1))
            r /= np.max(tools.absolute(r, axis=-1))
            plt.xlabel(r'$I_x$ (arb. units)')
            plt.ylabel(r'$I_y$ (arb. units)')
            ax.set_zlabel(r'$I_z$ (arb. units)')
            lim = 1.1
            ax.set_xlim(-lim, lim)
            ax.set_ylim(-lim, lim)
            ax.set_zlim(-lim, lim)
            ticks = [-1., -0.5, 0., 0.5, 1.]
            ax.set_xticks(ticks)
            ax.set_yticks(ticks)
            ax.set_zticks(ticks)
            ax.pbaspect = [1., 1., 1.]
            
            x_color = tools.absolute(r, axis=-1)
            cm = plt.cm.ScalarMappable(cmap='viridis')
            facecolors = cm.to_rgba(x_color)
            surf = ax.plot_surface(r[:, :, 0], r[:, :, 1], r[:, :, 2], rcount=80, ccount=80,
                                   facecolors=facecolors, antialiased=True, alpha=1)
            surf.set_clim([-1, 1])
            cb = plt.colorbar(cm, ax=ax, shrink=0.5)
            cb.set_label('Intensity (arb. units)')
            cb.set_ticks(np.arange(0., 1.1, 0.2))
            plt.gca().axis('off')
            plt.style.use('seaborn')
            plt.tight_layout()
        else:
            # import matplotlib
            font = {'family': 'Arial',
                    'size': 12}
            matplotlib.rc('font', **font)
            plt.figure(figsize=(3, 4.5))
            theta, phi = np.meshgrid(theta, phi, indexing='ij')
            ax = plt.gca()
            ax.set_aspect('equal')
            plt.xlabel(r'$\theta$ / rad')
            plt.ylabel(r'$\varphi$ / rad')
            cmesh = plt.pcolormesh(theta, phi, y[:, :], cmap='viridis')
            divider = make_axes_locatable(ax)
            cax = divider.append_axes('right', size='4%', pad=0.05)
            cbar = plt.colorbar(cmesh, cax=cax)
            cbar.set_label(label='Intensity (arb. units)')

        if save is not None:
            plt.savefig(save, dpi=300)
        if show:
            plt.show()
        plt.cla()
        plt.clf()
        plt.close()
        plt.style.use('default')
        return theta, phi, y

    def plot_mixed_distribution(self, n: int = 5, mode: str = None, show: bool = True, save: str = None):
        """
        :param n: 2 ** n + 1 samples for 'theta' and 2 ** (n + 1) + 1 samples for 'phi' are drawn,
         giving a total of (2 ** n + 1) * len(self.x) and (2 ** (n + 1) + 1) * len(self.x)
         points to draw for theta and phi, respectively.
        :param mode: Either '3d' for a surface plot or anything else for a color-mesh plot.
        :param save: The path where to save the plot. If None, the plot will not be saved.
        :param show: Whether to show the plot.
        :returns: None. Optionally, two plots are drawn for theta and phi as the y-axes and saved.
        """
        x = self.generate_x()
        theta = np.linspace(-np.pi / 2., np.pi / 2., 2 ** n + 1)
        phi = np.linspace(0., 2. * np.pi, 2 ** (n + 1) + 1)
        y = self.generate_y(x, theta, phi)
        x_theta, theta = np.meshgrid(x, theta, indexing='ij')
        x_phi, phi = np.meshgrid(x, phi, indexing='ij')
        plt.xlabel(r'$\nu - \nu_0$ / MHz')
        plt.ylabel(r'$\theta$ / rad')

        if mode is None:
            mode = '2d'
        if mode == '3d':
            fig = plt.figure(num=1, figsize=(16, 9), clear=True)
            ax = fig.add_subplot(111, projection='3d', label='Surface')
            ax.set_zlabel(r'$I$ / arb. units')
            ax.plot_surface(x_theta, theta, y[:, :, 0], rcount=200/2, ccount=200/2, cmap=plt.get_cmap('viridis'),
                            antialiased=True, alpha=1)
            if save is not None:
                plt.savefig(save, dpi=300)
                plt.close()
            if show:
                plt.show()
            fig = plt.figure(num=1, figsize=(16, 9), clear=True)
            ax = fig.add_subplot(111, projection='3d', label='Surface')
            plt.xlabel(r'$\nu - \nu_0$ / MHz')
            plt.ylabel(r'$\phi$ / rad')
            ax.set_zlabel(r'$I$ / arb. units')
            ax.plot_surface(x_phi, phi, y[:, 2 ** (n - 1), :], rcount=200/2, ccount=200/2, cmap=plt.get_cmap('viridis'),
                            antialiased=True, alpha=1)
            if save is not None:
                plt.savefig(save, dpi=300)
                plt.close()
            if show:
                plt.show()
        else:
            ax = plt.gca()
            cmesh = plt.pcolormesh(x_theta, theta, y[:, :, 0], cmap='viridis')
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="4%", pad=0.1)
            cbar = plt.colorbar(cmesh, cax=cax)
            cbar.set_label(label=r'$I$ / arb. units')
            if save is not None:
                plt.savefig(save, dpi=300)
                plt.close()
            if show:
                plt.show()
            plt.xlabel(r'$\nu - \nu_0$ / MHz')
            plt.ylabel(r'$\phi$ / rad')
            cmesh = plt.pcolormesh(x_phi,  phi, y[:, 2 ** (n - 1), :], cmap='viridis')
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="4%", pad=0.1)
            cbar = plt.colorbar(cmesh, cax=cax)
            cbar.set_label(label=r'$I$ / arb. units')
            if save is not None:
                plt.savefig(save, dpi=300)
                plt.close()
            if show:
                plt.show()

    def plot_setup(self):
        """
        :returns: None. Creates a 3D-plot of the physical situation, including the axis
         of the magnetic field (quantization axis), the real and imaginary axis of the polarization
         and the are which is covered by the detector geometry.
        """
        fig, ax = self.geometry.plot(show=False)
        x, y, z = np.zeros(1), np.zeros(1), np.zeros(1)

        # Laser

        # Polarization
        # Real
        u, v, w = self.polarization.x.real[:, 0], self.polarization.x.real[:, 1], self.polarization.x.real[:, 2]
        ax.quiver(x, y, z, u, v, w, zorder=3, color='b', length=1.,
                  arrow_length_ratio=0.1, label='Real(Pol.)')
        # Imag
        u, v, w = np.meshgrid(self.polarization.x.imag[:, 0], self.polarization.x.imag[:, 1],
                              self.polarization.x.imag[:, 2])
        ax.quiver(x, y, z, u, v, w, zorder=3, color='tab:orange', length=1.,
                  arrow_length_ratio=0.1, label='Imag(Pol.)')

        # B-field / Quantization axis
        u, v, w = np.meshgrid(self.e_r_b[0], self.e_r_b[1], self.e_r_b[2])
        ax.quiver(x, y, z, u, v, w, zorder=2, color='g', length=1.,
                  arrow_length_ratio=0.1, label='B-field /\nQuant. axis')

        # Velocity
        if self.v is not None:
            u, v, w = np.meshgrid(self.v[0], self.v[1], self.v[2])
            ax.quiver(x, y, z, u, v, w, zorder=10, color='r', length=1.,
                      arrow_length_ratio=0.1, label='Velocity')

        plt.legend()
        plt.show()
        plt.style.use('default')



# -*- coding: utf-8 -*-
"""
qspec.simulate
==============

Module for simulations of laser-atom interactions.
"""

from qspec.simulate import _simulate_cpp
from qspec.simulate import _simulate
from qspec.simulate._simulate_cpp import *
from qspec.simulate._simulate import *

__all__ = []
__all__.extend(_simulate_cpp.__all__)
__all__.extend(_simulate.__all__)
